"""Midea local FB device."""

import logging
from enum import StrEnum
from typing import Any, ClassVar

from midealocal.device import MideaDevice

from .message import MessageFBResponse, MessageQuery, MessageSet

_LOGGER = logging.getLogger(__name__)


class DeviceAttributes(StrEnum):
    """Midea FB device attributes."""

    power = "power"
    mode = "mode"
    heating_level = "heating_level"
    target_temperature = "target_temperature"
    current_temperature = "current_temperature"
    child_lock = "child_lock"


class MideaFBDevice(MideaDevice):
    """Midea FB device."""

    _modes: ClassVar[dict[int, str]] = {
        0x01: "Auto",
        0x02: "ECO",
        0x03: "Sleep",
        0x04: "Anti-freezing",
        0x05: "Comfort",
        0x06: "Constant-temperature",
        0x07: "Normal",
        0x08: "Fast-heating",
        0x10: "Standby",
    }

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,  # noqa: ARG002
    ) -> None:
        """Initialize Midea FB device."""
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0xFB,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.power: False,
                DeviceAttributes.mode: None,
                DeviceAttributes.heating_level: 0,
                DeviceAttributes.target_temperature: None,
                DeviceAttributes.current_temperature: None,
                DeviceAttributes.child_lock: False,
            },
        )

    @property
    def modes(self) -> list[str]:
        """Midea FB device modes."""
        return list(MideaFBDevice._modes.values())

    def build_query(self) -> list[MessageQuery]:
        """Midea FB device build query."""
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg: bytes) -> dict[str, Any]:
        """Midea FB device process message."""
        message = MessageFBResponse(msg)
        _LOGGER.debug("[%s] Received: %s", self.device_id, message)
        new_status = {}
        for status in self._attributes:
            if hasattr(message, str(status)):
                value = getattr(message, str(status))
                if status == DeviceAttributes.mode:
                    if value in MideaFBDevice._modes:
                        self._attributes[status] = MideaFBDevice._modes.get(value)
                    else:
                        self._attributes[status] = None
                else:
                    self._attributes[status] = value
                new_status[str(status)] = self._attributes[status]
        return new_status

    def set_attribute(self, attr: str, value: str | int | bool) -> None:
        """Midea FB device set attribute."""
        if attr == DeviceAttributes.mode:
            message = MessageSet(self._protocol_version, self.subtype)
            if value in MideaFBDevice._modes.values():
                message.mode = list(MideaFBDevice._modes.keys())[
                    list(MideaFBDevice._modes.values()).index(str(value))
                ]
        else:
            message = MessageSet(self._protocol_version, self.subtype)
            setattr(message, str(attr), value)
        self.build_send(message)

    def set_target_temperature(
        self,
        target_temperature: float,
        mode: int | None,  # noqa: ARG002
        zone: int | None = None,  # noqa: ARG002
    ) -> None:
        """Midea FB device set target temperature."""
        message = MessageSet(self._protocol_version, self.subtype)
        setattr(message, DeviceAttributes.target_temperature, target_temperature)
        self.build_send(message)


class MideaAppliance(MideaFBDevice):
    """Midea FB appliance."""



"""Midea AC message."""

from enum import IntEnum

from midealocal.const import MAX_BYTE_VALUE
from midealocal.crc8 import calculate
from midealocal.device import DeviceType
from midealocal.message import (
    BodyType,
    MessageBody,
    MessageRequest,
    MessageResponse,
    MessageType,
    NewProtocolMessageBody,
    SubBodyType,
)

BB_AC_MODES = [0, 3, 1, 2, 4, 5]
BB_MIN_BODY_LENGTH = 21
CONFORT_MODE_MIN_LENGTH = 16
CONFORT_MODE_C0_MIN_LENGTH = 23
ECO_MODE_MIN_SUBPROTOCOL_LENGTH = 27
FRESH_AIR_LENGTH = 2
FROST_PROTECT_C0_MIN_LENGTH = 22
INDIRECT_WIND_VALUE = 0x02
MAX_MSG_SERIAL_NUM = 254
SCREEN_DISPLAY_BYTE_CHECK = 0x07
SUB_PROTOCOL_BODY_TEMP_CHECK = 0x80
TEMP_DECIMAL_MIN_BODY_LENGTH = 20
TEMP_NEG_VALUE = 49
TIMER_MIN_SUBPROTOCOL_LENGTH = 27
XBB_SN8_BYTE_FLAG = 0x31
XC1_SUBBODY_TYPE_44 = 0x44
XC1_SUBBODY_TYPE_40 = 0x40


class PowerAnalysisMethod(IntEnum):
    """AC Power analysis method."""

    TYPE_1 = 1
    TYPE_2 = 2
    TYPE_3 = 3


class NewProtocolTags(IntEnum):
    """New protocol tags."""

    indoor_humidity = 0x0015
    screen_display = 0x0017
    breezeless = 0x0018
    prompt_tone = 0x001A
    indirect_wind = 0x0042
    fresh_air_1 = 0x0233
    fresh_air_2 = 0x004B


class MessageACBase(MessageRequest):
    """AC message base."""

    _message_serial = 0

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int,
    ) -> None:
        """Initialize AC message base."""
        super().__init__(
            device_type=DeviceType.AC,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )
        MessageACBase._message_serial += 1
        if MessageACBase._message_serial >= MAX_MSG_SERIAL_NUM:
            MessageACBase._message_serial = 1
        self._message_id = MessageACBase._message_serial

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError

    @property
    def body(self) -> bytearray:
        """AC message base body."""
        body = bytearray([self.body_type]) + self._body + bytearray([self._message_id])
        body.append(calculate(body))
        return body


class MessageQuery(MessageACBase):
    """AC message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize AC message query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=BodyType.X41,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray(
            [
                0x81,
                0x00,
                0xFF,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ],
        )


class MessagePowerQuery(MessageACBase):
    """AC message power query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize AC message power query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=BodyType.X41,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray([0x21, 0x01, 0x44, 0x00, 0x01])

    @property
    def body(self) -> bytearray:
        """AC message power query body."""
        body = bytearray([self.body_type]) + self._body
        body.append(calculate(body))
        return body


class MessageToggleDisplay(MessageACBase):
    """AC message toggle display."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize AC message toggle display."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=BodyType.X41,
        )
        self.prompt_tone = False

    @property
    def _body(self) -> bytearray:
        prompt_tone = 0x40 if self.prompt_tone else 0
        return bytearray(
            [
                0x02 | prompt_tone,
                0x00,
                0xFF,
                0x02,
                0x00,
                0x02,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ],
        )


class MessageNewProtocolQuery(MessageACBase):
    """AC message new protocol query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize AC message new protocol query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=BodyType.B1,
        )

    @property
    def _body(self) -> bytearray:
        query_params = [
            NewProtocolTags.indirect_wind,
            NewProtocolTags.breezeless,
            NewProtocolTags.indoor_humidity,
            NewProtocolTags.screen_display,
            NewProtocolTags.fresh_air_1,
            NewProtocolTags.fresh_air_2,
        ]

        _body = bytearray([len(query_params)])
        for param in query_params:
            _body.extend([param & 0xFF, param >> 8])
        return _body


class MessageSubProtocol(MessageACBase):
    """AC message sub protocol."""

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        subprotocol_query_type: int,
    ) -> None:
        """Initialize AC message sub protocol."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=0xAA,
        )
        self._subprotocol_query_type = subprotocol_query_type

    @property
    def _subprotocol_body(self) -> bytes:
        return bytes([])

    @property
    def body(self) -> bytearray:
        """AC message sub protocol body."""
        body = bytearray([self.body_type]) + self._body
        body.append(calculate(body))
        body.append(self.checksum(body))
        return body

    @property
    def _body(self) -> bytearray:
        _subprotocol_body = self._subprotocol_body
        _body = bytearray(
            [
                6
                + 2
                + (len(_subprotocol_body) if _subprotocol_body is not None else 0),
                0x00,
                0xFF,
                0xFF,
                self._subprotocol_query_type,
            ],
        )
        if _subprotocol_body is not None:
            _body.extend(_subprotocol_body)
        return _body


class MessageSubProtocolQuery(MessageSubProtocol):
    """AC message sub protocol query."""

    def __init__(self, protocol_version: int, subprotocol_query_type: int) -> None:
        """Initialize AC message sub protocol query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            subprotocol_query_type=subprotocol_query_type,
        )


class MessageSubProtocolSet(MessageSubProtocol):
    """AC message sub protocol set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize AC message sub protocol set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            subprotocol_query_type=0x20,
        )
        self.power = False
        self.mode = 0
        self.target_temperature = 20.0
        self.fan_speed = 102
        self.boost_mode = False
        self.aux_heating = False
        self.dry = False
        self.eco_mode = False
        self.sleep_mode = False
        self.sn8_flag = False
        self.timer = False
        self.prompt_tone = False

    @property
    def _subprotocol_body(self) -> bytes:
        power = 0x01 if self.power else 0
        dry = 0x10 if self.power and self.dry else 0
        boost_mode = 0x20 if self.boost_mode else 0
        aux_heating = 0x40 if self.aux_heating else 0x80
        sleep_mode = 0x80 if self.sleep_mode else 0
        try:
            mode = 0 if self.mode == 0 else BB_AC_MODES[self.mode] - 1
        except IndexError:
            mode = 2  # set Auto if invalid mode
        target_temperature = int(self.target_temperature * 2 + 30)
        water_model_temperature_set = int((self.target_temperature - 1) * 2 + 50)
        fan_speed = self.fan_speed
        eco = 0x40 if self.eco_mode else 0

        prompt_tone = 0x01 if self.prompt_tone else 0
        timer = 0x04 if (self.sn8_flag and self.timer) else 0
        return bytearray(
            [
                0x02 | boost_mode | power | dry,
                aux_heating,
                sleep_mode,
                0x00,
                0x00,
                mode,
                target_temperature,
                fan_speed,
                0x32,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x01,
                0x01,
                0x00,
                0x01,
                water_model_temperature_set,
                prompt_tone,
                target_temperature,
                0x32,
                0x66,
                0x00,
                eco | timer,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x08,
            ],
        )


class MessageGeneralSet(MessageACBase):
    """AC message general set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize AC message general set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x40,
        )
        self.power = False
        self.prompt_tone = True
        self.mode = 0
        self.target_temperature = 20.0
        self.fan_speed = 102
        self.swing_vertical = False
        self.swing_horizontal = False
        self.boost_mode = False
        self.smart_eye = False
        self.dry = False
        self.aux_heating = False
        self.eco_mode = False
        self.temp_fahrenheit = False
        self.sleep_mode = False
        self.natural_wind = False
        self.frost_protect = False
        self.comfort_mode = False

    @property
    def _body(self) -> bytearray:
        # Byte1, Power, prompt_tone
        power = 0x01 if self.power else 0
        prompt_tone = 0x40 if self.prompt_tone else 0
        # Byte2, mode target_temperature
        mode = (self.mode << 5) & 0xE0
        target_temperature = (int(self.target_temperature) & 0xF) | (
            0x10 if int(round(self.target_temperature * 2)) % 2 != 0 else 0
        )
        # Byte 3, fan_speed
        fan_speed = self.fan_speed & 0x7F
        # Byte 7, swing_mode
        swing_mode = (
            0x30
            | (0x0C if self.swing_vertical else 0)
            | (0x03 if self.swing_horizontal else 0)
        )
        # Byte 8, turbo
        boost_mode = 0x20 if self.boost_mode else 0
        # Byte 9 aux_heating eco_mode
        smart_eye = 0x01 if self.smart_eye else 0
        dry = 0x04 if self.dry else 0
        aux_heating = 0x08 if self.aux_heating else 0
        eco_mode = 0x80 if self.eco_mode else 0
        # Byte 10 temp_fahrenheit
        temp_fahrenheit = 0x04 if self.temp_fahrenheit else 0
        sleep_mode = 0x01 if self.sleep_mode else 0
        boost_mode_1 = 0x02 if self.boost_mode else 0
        # Byte 17 natural_wind
        natural_wind = 0x40 if self.natural_wind else 0
        # Byte 21 frost_protect
        frost_protect = 0x80 if self.frost_protect else 0
        # Byte 22 comfort_mode
        comfort_mode = 0x01 if self.comfort_mode else 0

        return bytearray(
            [
                power | prompt_tone,
                mode | target_temperature,
                fan_speed,
                0x00,
                0x00,
                0x00,
                swing_mode,
                boost_mode,
                smart_eye | dry | aux_heating | eco_mode,
                temp_fahrenheit | sleep_mode | boost_mode_1,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                natural_wind,
                0x00,
                0x00,
                0x00,
                frost_protect,
                comfort_mode,
            ],
        )


class MessageNewProtocolSet(MessageACBase):
    """AC message new protocol set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize AC message new protocol set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0xB0,
        )
        self.indirect_wind: bytes | None = None
        self.prompt_tone: bytes | None = None
        self.breezeless: bytes | None = None
        self.screen_display_alternate: bytes | None = None
        self.fresh_air_1: bytes | None = None
        self.fresh_air_2: bytes | None = None

    @property
    def _body(self) -> bytearray:
        pack_count = 0
        payload = bytearray([0x00])
        if self.breezeless is not None:
            pack_count += 1
            payload.extend(
                NewProtocolMessageBody.pack(
                    param=NewProtocolTags.breezeless,
                    value=bytearray([0x01 if self.breezeless else 0x00]),
                ),
            )
        if self.indirect_wind is not None:
            pack_count += 1
            payload.extend(
                NewProtocolMessageBody.pack(
                    param=NewProtocolTags.indirect_wind,
                    value=bytearray([0x02 if self.indirect_wind else 0x01]),
                ),
            )
        if self.prompt_tone is not None:
            pack_count += 1
            payload.extend(
                NewProtocolMessageBody.pack(
                    param=NewProtocolTags.prompt_tone,
                    value=bytearray([0x01 if self.prompt_tone else 0x00]),
                ),
            )
        if self.screen_display_alternate is not None:
            pack_count += 1
            payload.extend(
                NewProtocolMessageBody.pack(
                    param=NewProtocolTags.screen_display,
                    value=bytearray([0x64 if self.screen_display_alternate else 0x00]),
                ),
            )
        if self.fresh_air_1 is not None and len(self.fresh_air_1) == FRESH_AIR_LENGTH:
            pack_count += 1
            fresh_air_power = 2 if self.fresh_air_1[0] > 0 else 1
            fresh_air_fan_speed = self.fresh_air_1[1]
            payload.extend(
                NewProtocolMessageBody.pack(
                    param=NewProtocolTags.fresh_air_1,
                    value=bytearray(
                        [
                            fresh_air_power,
                            fresh_air_fan_speed,
                            0x00,
                            0x00,
                            0x00,
                            0x00,
                            0x00,
                            0x00,
                            0x00,
                            0x00,
                        ],
                    ),
                ),
            )
        if self.fresh_air_2 is not None and len(self.fresh_air_2) == FRESH_AIR_LENGTH:
            pack_count += 1
            fresh_air_power = 1 if self.fresh_air_2[0] > 0 else 0
            fresh_air_fan_speed = self.fresh_air_2[1]
            payload.extend(
                NewProtocolMessageBody.pack(
                    param=NewProtocolTags.fresh_air_2,
                    value=bytearray([fresh_air_power, fresh_air_fan_speed, 0xFF]),
                ),
            )
        payload[0] = pack_count
        return payload


class XA0MessageBody(MessageBody):
    """AC A0 message body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize AC A0 message body."""
        super().__init__(body)
        self.power = (body[1] & 0x1) > 0
        self.target_temperature = (
            ((body[1] & 0x3E) >> 1) - 4 + 16.0 + (0.5 if body[1] & 0x40 > 0 else 0.0)
        )
        self.mode = (body[2] & 0xE0) >> 5
        self.fan_speed = body[3] & 0x7F
        self.swing_vertical = (body[7] & 0xC) > 0
        self.swing_horizontal = (body[7] & 0x3) > 0
        self.boost_mode = ((body[8] & 0x20) > 0) or ((body[10] & 0x2) > 0)
        self.smart_eye = (body[9] & 0x01) > 0
        self.dry = (body[9] & 0x04) > 0
        self.aux_heating = (body[9] & 0x08) > 0
        self.eco_mode = (body[9] & 0x10) > 0
        self.sleep_mode = (body[10] & 0x01) > 0
        self.natural_wind = (body[10] & 0x40) > 0
        self.full_dust = (body[13] & 0x20) > 0
        self.comfort_mode = (
            (body[14] & 0x1) > 0 if len(body) > CONFORT_MODE_MIN_LENGTH else False
        )


class XA1MessageBody(MessageBody):
    """AC A1 message body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize AC A1 message body."""
        super().__init__(body)
        if body[13] != MAX_BYTE_VALUE:
            temp_integer = int((body[13] - 50) / 2)
            temp_decimal = (
                ((body[18] & 0xF) * 0.1)
                if len(body) > TEMP_DECIMAL_MIN_BODY_LENGTH
                else 0
            )
            if body[13] > TEMP_NEG_VALUE:
                self.indoor_temperature = temp_integer + temp_decimal
            else:
                self.indoor_temperature = temp_integer - temp_decimal
        if body[14] == MAX_BYTE_VALUE:
            self.outdoor_temperature = None
        else:
            temp_integer = int((body[14] - 50) / 2)
            temp_decimal = (
                (((body[18] & 0xF0) >> 4) * 0.1)
                if len(body) > TEMP_DECIMAL_MIN_BODY_LENGTH
                else 0
            )
            if body[14] > TEMP_NEG_VALUE:
                self.outdoor_temperature = temp_integer + temp_decimal
            else:
                self.outdoor_temperature = temp_integer - temp_decimal
        self.indoor_humidity = body[17]


class XBXMessageBody(NewProtocolMessageBody):
    """AC BX message body."""

    def __init__(self, body: bytearray, bt: int) -> None:
        """Initialize AC BX message body."""
        super().__init__(body, bt)
        params = self.parse()
        if NewProtocolTags.indirect_wind in params:
            self.indirect_wind = (
                params[NewProtocolTags.indirect_wind][0] == INDIRECT_WIND_VALUE
            )
        if NewProtocolTags.indoor_humidity in params:
            self.indoor_humidity = params[NewProtocolTags.indoor_humidity][0]
        if NewProtocolTags.breezeless in params:
            self.breezeless = params[NewProtocolTags.breezeless][0] == 1
        if NewProtocolTags.screen_display in params:
            self.screen_display_alternate = (
                params[NewProtocolTags.screen_display][0] > 0
            )
            self.screen_display_new = True
        if NewProtocolTags.fresh_air_1 in params:
            self.fresh_air_1 = True
            data = params[NewProtocolTags.fresh_air_1]
            self.fresh_air_power = data[0] == INDIRECT_WIND_VALUE
            self.fresh_air_fan_speed = data[1]
        if NewProtocolTags.fresh_air_2 in params:
            self.fresh_air_2 = True
            data = params[NewProtocolTags.fresh_air_2]
            self.fresh_air_power = data[0] > 0
            self.fresh_air_fan_speed = data[1]


class XC0MessageBody(MessageBody):
    """AC C0 message body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize AC C0 message body."""
        super().__init__(body)
        self.power = (body[1] & 0x1) > 0
        self.mode = (body[2] & 0xE0) >> 5
        self.target_temperature = (
            (body[2] & 0x0F) + 16.0 + (0.5 if body[0x02] & 0x10 > 0 else 0.0)
        )
        self.fan_speed = body[3] & 0x7F
        self.swing_vertical = (body[7] & 0x0C) > 0
        self.swing_horizontal = (body[7] & 0x03) > 0
        self.boost_mode = ((body[8] & 0x20) > 0) or ((body[10] & 0x2) > 0)
        self.smart_eye = (body[8] & 0x40) > 0
        self.natural_wind = (body[9] & 0x2) > 0
        self.dry = (body[9] & 0x4) > 0
        self.eco_mode = (body[9] & 0x10) > 0
        self.aux_heating = (body[9] & 0x08) > 0
        self.temp_fahrenheit = (body[10] & 0x04) > 0
        self.sleep_mode = (body[10] & 0x01) > 0
        if body[11] != MAX_BYTE_VALUE:
            temp_integer = int((body[11] - 50) / 2)
            temp_decimal = (body[15] & 0x0F) * 0.1
            if body[11] > TEMP_NEG_VALUE:
                self.indoor_temperature = temp_integer + temp_decimal
            else:
                self.indoor_temperature = temp_integer - temp_decimal
        if body[12] == MAX_BYTE_VALUE:
            self.outdoor_temperature = None
        else:
            temp_integer = int((body[12] - 50) / 2)
            temp_decimal = ((body[15] & 0xF0) >> 4) * 0.1
            if body[12] > TEMP_NEG_VALUE:
                self.outdoor_temperature = temp_integer + temp_decimal
            else:
                self.outdoor_temperature = temp_integer - temp_decimal
        self.full_dust = (body[13] & 0x20) > 0
        self.screen_display = (
            (body[14] >> 4 & 0x7) != SCREEN_DISPLAY_BYTE_CHECK
        ) and self.power
        self.frost_protect = (
            (body[21] & 0x80) > 0 if len(body) >= FROST_PROTECT_C0_MIN_LENGTH else False
        )
        self.comfort_mode = (
            (body[22] & 0x1) > 0 if len(body) >= CONFORT_MODE_C0_MIN_LENGTH else False
        )


class XC1MessageBody(MessageBody):
    """AC C1 message body."""

    def __init__(self, body: bytearray, analysis_method: int = 3) -> None:
        """Initialize AC C1 message body."""
        super().__init__(body)
        if body[3] == XC1_SUBBODY_TYPE_44:
            self.total_energy_consumption = XC1MessageBody.parse_consumption(
                analysis_method,
                body[4],
                body[5],
                body[6],
                body[7],
            )
            self.current_energy_consumption = XC1MessageBody.parse_consumption(
                analysis_method,
                body[12],
                body[13],
                body[14],
                body[15],
            )
            self.realtime_power = XC1MessageBody.parse_power(
                analysis_method,
                body[16],
                body[17],
                body[18],
            )
        elif body[3] == XC1_SUBBODY_TYPE_40:
            pass

    @staticmethod
    def parse_value(byte: int) -> int:
        """AC C1 message body parse value."""
        return (byte >> 4) * 10 + (byte & 0x0F)

    @staticmethod
    def parse_power(analysis_method: int, byte1: int, byte2: int, byte3: int) -> float:
        """AC C1 message body parse power."""
        if analysis_method == PowerAnalysisMethod.TYPE_1:
            return (
                float(
                    XC1MessageBody.parse_value(byte1) * 10000
                    + XC1MessageBody.parse_value(byte2) * 100
                    + XC1MessageBody.parse_value(byte3),
                )
                / 10
            )
        if analysis_method == PowerAnalysisMethod.TYPE_2:
            return float((byte1 << 16) + (byte2 << 8) + byte3) / 10
        return float(byte1 * 10000 + byte2 * 100 + byte3) / 10

    @staticmethod
    def parse_consumption(
        analysis_method: int,
        byte1: int,
        byte2: int,
        byte3: int,
        byte4: int,
    ) -> float:
        """AC C1 message body parse consumption."""
        if analysis_method == PowerAnalysisMethod.TYPE_1:
            return (
                float(
                    XC1MessageBody.parse_value(byte1) * 1000000
                    + XC1MessageBody.parse_value(byte2) * 10000
                    + XC1MessageBody.parse_value(byte3) * 100
                    + XC1MessageBody.parse_value(byte4),
                )
                / 100
            )
        if analysis_method == PowerAnalysisMethod.TYPE_2:
            return float((byte1 << 32) + (byte2 << 16) + (byte3 << 8) + byte4) / 10
        return float(byte1 * 1000000 + byte2 * 10000 + byte3 * 100 + byte4) / 100


class XBBMessageBody(MessageBody):
    """AC BB message body."""

    def __init__(self, body: bytearray) -> None:
        """Initializ AC BB message body."""
        super().__init__(body)
        subprotocol_head = body[:6]
        subprotocol_body = body[6:]
        data_type = subprotocol_head[-1]
        subprotocol_body_len = len(subprotocol_body)
        if data_type in (SubBodyType.X11, SubBodyType.X20):
            self.power = (subprotocol_body[0] & 0x1) > 0
            self.dry = (subprotocol_body[0] & 0x10) > 0
            self.boost_mode = (subprotocol_body[0] & 0x20) > 0
            self.aux_heating = (subprotocol_body[1] & 0x40) > 0
            self.sleep_mode = (subprotocol_body[2] & 0x80) > 0
            try:
                self.mode = BB_AC_MODES.index(subprotocol_body[5] + 1)
            except ValueError:
                self.mode = 0
            self.target_temperature = (subprotocol_body[6] - 30) / 2
            self.fan_speed = subprotocol_body[7]
            self.timer = (
                (subprotocol_body[25] & 0x04) > 0
                if subprotocol_body_len > TIMER_MIN_SUBPROTOCOL_LENGTH
                else False
            )
            self.eco_mode = (
                (subprotocol_body[25] & 0x40) > 0
                if subprotocol_body_len > ECO_MODE_MIN_SUBPROTOCOL_LENGTH
                else False
            )
        elif data_type == SubBodyType.X10:
            if subprotocol_body[8] & 0x80 == SUB_PROTOCOL_BODY_TEMP_CHECK:
                self.indoor_temperature = (
                    0 - (~(subprotocol_body[7] + subprotocol_body[8] * 256) + 1)
                    & 0xFFFF
                ) / 100
            else:
                self.indoor_temperature = (
                    subprotocol_body[7] + subprotocol_body[8] * 256
                ) / 100
            self.indoor_humidity = subprotocol_body[30]
            self.sn8_flag = subprotocol_body[80] == XBB_SN8_BYTE_FLAG
        elif data_type == SubBodyType.X12:
            pass
        elif data_type == SubBodyType.X30:
            if subprotocol_body[6] & 0x80 == SUB_PROTOCOL_BODY_TEMP_CHECK:
                self.outdoor_temperature = (
                    0 - (~(subprotocol_body[5] + subprotocol_body[6] * 256) + 1)
                    & 0xFFFF
                ) / 100
            else:
                self.outdoor_temperature = (
                    subprotocol_body[5] + subprotocol_body[6] * 256
                ) / 100
        elif data_type in (SubBodyType.X13, SubBodyType.X21):
            pass


class MessageACResponse(MessageResponse):
    """AC message response."""

    def __init__(self, message: bytearray, power_analysis_method: int = 3) -> None:
        """Initialize AC message response."""
        super().__init__(message)
        if self.message_type == MessageType.notify2 and self.body_type == BodyType.A0:
            self.set_body(XA0MessageBody(super().body))
        elif self.message_type == MessageType.notify1 and self.body_type == BodyType.A1:
            self.set_body(XA1MessageBody(super().body))
        elif self.message_type in [
            MessageType.query,
            MessageType.set,
            MessageType.notify2,
        ] and self.body_type in [BodyType.B0, BodyType.B1, BodyType.B5]:
            self.set_body(XBXMessageBody(super().body, self.body_type))
        elif (
            self.message_type in [MessageType.query, MessageType.set]
            and self.body_type == BodyType.C0
        ):
            self.set_body(XC0MessageBody(super().body))
        elif self.message_type == MessageType.query and self.body_type == BodyType.C1:
            self.set_body(XC1MessageBody(super().body, power_analysis_method))
        elif (
            self.message_type
            in [MessageType.set, MessageType.query, MessageType.notify2]
            and self.body_type == BodyType.BB
            and len(super().body) >= BB_MIN_BODY_LENGTH
        ):
            self.used_subprotocol = True
            self.set_body(XBBMessageBody(super().body))
        self.set_attr()



"""Midea local E6 message."""

from midealocal.message import (
    NONE_VALUE,
    MessageBody,
    MessageRequest,
    MessageResponse,
    MessageType,
)


class MessageE6Base(MessageRequest):
    """E6 message base."""

    def __init__(self, protocol_version: int, message_type: int) -> None:
        """Initialize E6 message base."""
        super().__init__(
            device_type=0xE6,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=NONE_VALUE,
        )

    @property
    def body(self) -> bytearray:
        """E6 message base body."""
        return self._body

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError


class MessageQuery(MessageE6Base):
    """E6 message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize E6 message query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray([0x01, 0x01] + [0] * 28)


class MessageSet(MessageE6Base):
    """E6 message set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize E6 message set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
        )
        self.main_power: bool | None = None
        self.heating_temperature: int | None = None
        self.bathing_temperature: int | None = None
        self.heating_power: bool | None = None

    @property
    def _body(self) -> bytearray:
        body: list[int] = []
        if self.main_power is not None:
            main_power = 0x01 if self.main_power else 0x02
            body = [main_power, 0x01]
        elif self.heating_temperature is not None:
            body = [0x04, 0x13, self.heating_temperature]
        elif self.bathing_temperature is not None:
            body = [0x04, 0x12, self.bathing_temperature]
        elif self.heating_power is not None:
            heating_power = 0x01 if self.heating_power else 0x02
            body = [0x04, 0x01, heating_power]
        body_len = len(body)
        return bytearray(body + [0] * (30 - body_len))


class E6GeneralMessageBody(MessageBody):
    """E6 message general body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize E6 message general body."""
        super().__init__(body)
        self.main_power = (body[2] & 0x04) > 0
        self.heating_working = (body[2] & 0x10) > 0
        self.bathing_working = (body[2] & 0x20) > 0
        self.heating_power = (body[4] & 0x01) > 0
        self.min_temperature = [body[16], body[11]]
        self.max_temperature = [body[15], body[10]]
        self.heating_temperature = body[17]
        self.bathing_temperature = body[12]
        self.heating_leaving_temperature = body[14]
        self.bathing_leaving_temperature = body[8]


class MessageE6Response(MessageResponse):
    """E6 message response."""

    def __init__(self, message: bytes) -> None:
        """Initialize E6 message response."""
        super().__init__(bytearray(message))
        self.set_body(E6GeneralMessageBody(super().body))
        self.set_attr()



"""Midea local C3 message."""

from midealocal.message import (
    BodyType,
    MessageBody,
    MessageRequest,
    MessageResponse,
    MessageType,
)

TEMP_NEG_VALUE = 127


class MessageC3Base(MessageRequest):
    """C3 message base."""

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int,
    ) -> None:
        """Initialize C3 message base."""
        super().__init__(
            device_type=0xC3,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError


class MessageQuery(MessageC3Base):
    """C3 message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize C3 message query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x01,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray([])


class MessageSet(MessageC3Base):
    """C3 message set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize C3 message set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x01,
        )
        self.zone1_power = False
        self.zone2_power = False
        self.dhw_power = False
        self.mode = 0
        self.zone_target_temp = [25.0, 25.0]
        self.dhw_target_temp = 40
        self.room_target_temp = 25.0
        self.zone1_curve = False
        self.zone2_curve = False
        self.disinfect = False
        self.fast_dhw = False
        self.tbh = False

    @property
    def _body(self) -> bytearray:
        # Byte 1
        zone1_power = 0x01 if self.zone1_power else 0x00
        zone2_power = 0x02 if self.zone2_power else 0x00
        dhw_power = 0x04 if self.dhw_power else 0x00
        # Byte 7
        zone1_curve = 0x01 if self.zone1_curve else 0x00
        zone2_curve = 0x02 if self.zone2_curve else 0x00
        disinfect = 0x04 if self.disinfect or self.tbh else 0x00
        fast_dhw = 0x08 if self.fast_dhw else 0x00
        room_target_temp = int(self.room_target_temp * 2)
        zone1_target_temp = int(self.zone_target_temp[0])
        zone2_target_temp = int(self.zone_target_temp[1])
        dhw_target_temp = int(self.dhw_target_temp)
        return bytearray(
            [
                zone1_power | zone2_power | dhw_power,
                self.mode,
                zone1_target_temp,
                zone2_target_temp,
                dhw_target_temp,
                room_target_temp,
                zone1_curve | zone2_curve | disinfect | fast_dhw,
            ],
        )


class MessageSetSilent(MessageC3Base):
    """C3 message set silent."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize C3 message set silent."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x05,
        )
        self.silent_mode = False
        self.super_silent = False

    @property
    def _body(self) -> bytearray:
        silent_mode = 0x01 if self.silent_mode else 0
        super_silent = 0x02 if self.super_silent else 0

        return bytearray(
            [
                silent_mode | super_silent,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ],
        )


class MessageSetECO(MessageC3Base):
    """C3 message set eco."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize C3 message set eco."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x07,
        )
        self.eco_mode = False

    @property
    def _body(self) -> bytearray:
        eco_mode = 0x01 if self.eco_mode else 0

        return bytearray([eco_mode, 0x00, 0x00, 0x00, 0x00, 0x00])


class C3MessageBody(MessageBody):
    """C3 message body."""

    def __init__(self, body: bytearray, data_offset: int = 0) -> None:
        """Initialize C3 message body."""
        super().__init__(body)
        self.zone1_power = body[data_offset + 0] & 0x01 > 0
        self.zone2_power = body[data_offset + 0] & 0x02 > 0
        self.dhw_power = body[data_offset + 0] & 0x04 > 0
        self.zone1_curve = body[data_offset + 0] & 0x08 > 0
        self.zone2_curve = body[data_offset + 0] & 0x10 > 0
        self.disinfect = body[data_offset + 0] & 0x20 > 0
        self.tbh = body[data_offset + 0] & 0x20 > 0
        self.fast_dhw = body[data_offset + 0] & 0x40 > 0
        self.zone_temp_type = [
            body[data_offset + 1] & 0x10 > 0,
            body[data_offset + 1] & 0x20 > 0,
        ]
        self.silent_mode = body[data_offset + 2] & 0x02 > 0
        self.eco_mode = body[data_offset + 2] & 0x08 > 0
        self.mode = body[data_offset + 3]
        self.mode_auto = body[data_offset + 4]
        self.zone_target_temp = [body[data_offset + 5], body[data_offset + 6]]
        self.dhw_target_temp = body[data_offset + 7]
        self.room_target_temp = body[data_offset + 8] / 2
        self.zone_heating_temp_max = [body[data_offset + 9], body[data_offset + 13]]
        self.zone_heating_temp_min = [body[data_offset + 10], body[data_offset + 14]]
        self.zone_cooling_temp_max = [body[data_offset + 11], body[data_offset + 15]]
        self.zone_cooling_temp_min = [body[data_offset + 12], body[data_offset + 16]]
        self.room_temp_max = body[data_offset + 17] / 2
        self.room_temp_min = body[data_offset + 18] / 2
        self.dhw_temp_max = body[data_offset + 19]
        self.dhw_temp_min = body[data_offset + 20]
        self.tank_actual_temperature = body[data_offset + 21]
        self.error_code = body[data_offset + 22]


class C3Notify1MessageBody(MessageBody):
    """C3 notify1 message body."""

    def __init__(self, body: bytearray, data_offset: int = 0) -> None:
        """Initialize C3 notify1 message body."""
        super().__init__(body)
        status_byte = body[data_offset]
        self.status_tbh = (status_byte & 0x08) > 0
        self.status_dhw = (status_byte & 0x04) > 0
        self.status_ibh = (status_byte & 0x02) > 0
        self.status_heating = (status_byte & 0x01) > 0

        self.total_energy_consumption = (
            (body[data_offset + 1] << 32)
            + (body[data_offset + 2] << 16)
            + (body[data_offset + 3] << 8)
            + (body[data_offset + 4])
        )

        self.total_produced_energy = (
            (body[data_offset + 5] << 32)
            + (body[data_offset + 6] << 16)
            + (body[data_offset + 7] << 8)
            + (body[data_offset + 8])
        )
        base_value = body[data_offset + 9]
        self.outdoor_temperature = (
            (base_value - 256) if base_value > TEMP_NEG_VALUE else base_value
        )


class MessageC3Response(MessageResponse):
    """C3 message response."""

    def __init__(self, message: bytes) -> None:
        """Initialize C3 message response."""
        super().__init__(bytearray(message))
        if (
            self.message_type
            in [MessageType.set, MessageType.notify1, MessageType.query]
            and self.body_type == BodyType.X01
        ) or self.message_type == MessageType.notify2:
            self.set_body(C3MessageBody(super().body, data_offset=1))
        elif (
            self.message_type == MessageType.notify1 and self.body_type == BodyType.X04
        ):
            self.set_body(C3Notify1MessageBody(super().body, data_offset=1))
        self.set_attr()



"""Midea local AC device."""

import json
import logging
from enum import StrEnum
from typing import Any, ClassVar

from midealocal.device import MideaDevice

from .message import (
    MessageACResponse,
    MessageGeneralSet,
    MessageNewProtocolQuery,
    MessageNewProtocolSet,
    MessagePowerQuery,
    MessageQuery,
    MessageSubProtocolQuery,
    MessageSubProtocolSet,
    MessageToggleDisplay,
)

_LOGGER = logging.getLogger(__name__)


class DeviceAttributes(StrEnum):
    """Midea AC device attributes."""

    prompt_tone = "prompt_tone"
    power = "power"
    mode = "mode"
    target_temperature = "target_temperature"
    fan_speed = "fan_speed"
    swing_vertical = "swing_vertical"
    swing_horizontal = "swing_horizontal"
    boost_mode = "boost_mode"
    smart_eye = "smart_eye"
    dry = "dry"
    eco_mode = "eco_mode"
    aux_heating = "aux_heating"
    sleep_mode = "sleep_mode"
    natural_wind = "natural_wind"
    temp_fahrenheit = "temp_fahrenheit"
    screen_display = "screen_display"
    screen_display_alternate = "screen_display_alternate"
    full_dust = "full_dust"
    frost_protect = "frost_protect"
    comfort_mode = "comfort_mode"
    indoor_temperature = "indoor_temperature"
    outdoor_temperature = "outdoor_temperature"
    indirect_wind = "indirect_wind"
    indoor_humidity = "indoor_humidity"
    breezeless = "breezeless"
    fresh_air_power = "fresh_air_power"
    fresh_air_fan_speed = "fresh_air_fan_speed"
    fresh_air_mode = "fresh_air_mode"
    fresh_air_1 = "fresh_air_1"
    fresh_air_2 = "fresh_air_2"
    total_energy_consumption = "total_energy_consumption"
    current_energy_consumption = "current_energy_consumption"
    realtime_power = "realtime_power"


class MideaACDevice(MideaDevice):
    """Midea AC device."""

    _fresh_air_fan_speeds: ClassVar[dict[int, str]] = {
        0: "Off",
        20: "Silent",
        40: "Low",
        60: "Medium",
        80: "High",
        100: "Full",
    }

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,
    ) -> None:
        """Initialize Midea AC device."""
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0xAC,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.prompt_tone: True,
                DeviceAttributes.power: False,
                DeviceAttributes.mode: 0,
                DeviceAttributes.target_temperature: 24.0,
                DeviceAttributes.fan_speed: 102,
                DeviceAttributes.swing_vertical: False,
                DeviceAttributes.swing_horizontal: False,
                DeviceAttributes.smart_eye: False,
                DeviceAttributes.dry: False,
                DeviceAttributes.aux_heating: False,
                DeviceAttributes.boost_mode: False,
                DeviceAttributes.sleep_mode: False,
                DeviceAttributes.frost_protect: False,
                DeviceAttributes.comfort_mode: False,
                DeviceAttributes.eco_mode: False,
                DeviceAttributes.natural_wind: False,
                DeviceAttributes.temp_fahrenheit: False,
                DeviceAttributes.screen_display: False,
                DeviceAttributes.screen_display_alternate: False,
                DeviceAttributes.full_dust: False,
                DeviceAttributes.indoor_temperature: None,
                DeviceAttributes.outdoor_temperature: None,
                DeviceAttributes.indirect_wind: False,
                DeviceAttributes.indoor_humidity: None,
                DeviceAttributes.breezeless: False,
                DeviceAttributes.total_energy_consumption: None,
                DeviceAttributes.current_energy_consumption: None,
                DeviceAttributes.realtime_power: None,
                DeviceAttributes.fresh_air_power: False,
                DeviceAttributes.fresh_air_fan_speed: 0,
                DeviceAttributes.fresh_air_mode: None,
                DeviceAttributes.fresh_air_1: None,
                DeviceAttributes.fresh_air_2: None,
            },
        )
        self._fresh_air_version: DeviceAttributes | None = None
        self._default_temperature_step: float = 0.5
        self._temperature_step: float = 0.5
        self._used_subprotocol: bool = False
        self._bb_sn8_flag: bool = False
        self._bb_timer: bool = False
        self._power_analysis_method: int = 1
        self._default_power_analysis_method: int = 1
        self.set_customize(customize)

    @property
    def temperature_step(self) -> float | None:
        """Midea AC device temperature step."""
        return self._temperature_step

    @property
    def fresh_air_fan_speeds(self) -> list[str]:
        """Midea AC device fresh air fan speeds."""
        return list(MideaACDevice._fresh_air_fan_speeds.values())

    def build_query(
        self,
    ) -> list[
        MessageSubProtocolQuery
        | MessageQuery
        | MessageNewProtocolQuery
        | MessagePowerQuery
    ]:
        """Midea AC device build query."""
        if self._used_subprotocol:
            return [
                MessageSubProtocolQuery(self._protocol_version, 0x10),
                MessageSubProtocolQuery(self._protocol_version, 0x11),
                MessageSubProtocolQuery(self._protocol_version, 0x30),
            ]
        return [
            MessageQuery(self._protocol_version),
            MessageNewProtocolQuery(self._protocol_version),
            MessagePowerQuery(self._protocol_version),
        ]

    def process_message(self, msg: bytes) -> dict[str, Any]:
        """Midea AC device process message."""
        message = MessageACResponse(bytearray(msg), self._power_analysis_method)
        _LOGGER.debug("[%s] Received: %s", self.device_id, message)
        new_status = {}
        has_fresh_air = False
        if hasattr(message, "used_subprotocol"):
            self._used_subprotocol = True
            if hasattr(message, "sn8_flag"):
                self._bb_sn8_flag = message.sn8_flag
            if hasattr(message, "timer"):
                self._bb_timer = message.timer
        for status in self._attributes:
            if hasattr(message, str(status)):
                value = getattr(message, str(status))
                if status == DeviceAttributes.fresh_air_power:
                    has_fresh_air = True
                self._attributes[status] = value
                new_status[str(status)] = self._attributes[status]
        if has_fresh_air:
            if self._attributes[DeviceAttributes.fresh_air_power]:
                for k, v in MideaACDevice._fresh_air_fan_speeds.items():
                    if self._attributes[DeviceAttributes.fresh_air_fan_speed] < k:
                        break
                    self._attributes[DeviceAttributes.fresh_air_mode] = v
            else:
                self._attributes[DeviceAttributes.fresh_air_mode] = "Off"
            new_status[DeviceAttributes.fresh_air_mode.value] = self._attributes[
                DeviceAttributes.fresh_air_mode
            ]
        if not self._attributes[DeviceAttributes.power] or (
            DeviceAttributes.swing_vertical in new_status
            and self._attributes[DeviceAttributes.swing_vertical]
        ):
            self._attributes[DeviceAttributes.indirect_wind] = False
            new_status[DeviceAttributes.indirect_wind.value] = False
        if not self._attributes[DeviceAttributes.power]:
            self._attributes[DeviceAttributes.screen_display] = False
            new_status[DeviceAttributes.screen_display.value] = False
        if self._attributes[DeviceAttributes.fresh_air_1] is not None:
            self._fresh_air_version = DeviceAttributes.fresh_air_1
        elif self._attributes[DeviceAttributes.fresh_air_2] is not None:
            self._fresh_air_version = DeviceAttributes.fresh_air_2
        return new_status

    def make_message_set(self) -> MessageGeneralSet:
        """Midea AC device make message set."""
        message = MessageGeneralSet(self._protocol_version)
        message.power = self._attributes[DeviceAttributes.power]
        message.prompt_tone = self._attributes[DeviceAttributes.prompt_tone]
        message.mode = self._attributes[DeviceAttributes.mode]
        message.target_temperature = self._attributes[
            DeviceAttributes.target_temperature
        ]
        message.fan_speed = self._attributes[DeviceAttributes.fan_speed]
        message.swing_vertical = self._attributes[DeviceAttributes.swing_vertical]
        message.swing_horizontal = self._attributes[DeviceAttributes.swing_horizontal]
        message.boost_mode = self._attributes[DeviceAttributes.boost_mode]
        message.smart_eye = self._attributes[DeviceAttributes.smart_eye]
        message.dry = self._attributes[DeviceAttributes.dry]
        message.eco_mode = self._attributes[DeviceAttributes.eco_mode]
        message.aux_heating = self._attributes[DeviceAttributes.aux_heating]
        message.sleep_mode = self._attributes[DeviceAttributes.sleep_mode]
        message.natural_wind = self._attributes[DeviceAttributes.natural_wind]
        message.temp_fahrenheit = self._attributes[DeviceAttributes.temp_fahrenheit]
        message.frost_protect = self._attributes[DeviceAttributes.frost_protect]
        message.comfort_mode = self._attributes[DeviceAttributes.comfort_mode]
        return message

    def make_subptotocol_message_set(self) -> MessageSubProtocolSet:
        """Midea AC device make subprotocol message set."""
        message = MessageSubProtocolSet(self._protocol_version)
        message.power = self._attributes[DeviceAttributes.power]
        message.prompt_tone = self._attributes[DeviceAttributes.prompt_tone]
        message.aux_heating = self._attributes[DeviceAttributes.aux_heating]
        message.mode = self._attributes[DeviceAttributes.mode]
        message.target_temperature = self._attributes[
            DeviceAttributes.target_temperature
        ]
        message.fan_speed = self._attributes[DeviceAttributes.fan_speed]
        message.boost_mode = self._attributes[DeviceAttributes.boost_mode]
        message.dry = self._attributes[DeviceAttributes.dry]
        message.eco_mode = self._attributes[DeviceAttributes.eco_mode]
        message.sleep_mode = self._attributes[DeviceAttributes.sleep_mode]
        message.sn8_flag = self._bb_sn8_flag
        message.timer = self._bb_timer
        return message

    def make_message_uniq_set(self) -> MessageSubProtocolSet | MessageGeneralSet:
        """Midea AC device make message unique set."""
        message: MessageSubProtocolSet | MessageGeneralSet
        if self._used_subprotocol:
            message = self.make_subptotocol_message_set()
        else:
            message = self.make_message_set()
        return message

    def set_attribute(self, attr: str, value: bool | int | str) -> None:
        """Midea AC device set attribute."""
        # if nat a sensor
        message: (
            MessageToggleDisplay | MessageNewProtocolSet | MessageGeneralSet | None
        ) = None
        if attr not in [
            DeviceAttributes.indoor_temperature,
            DeviceAttributes.outdoor_temperature,
            DeviceAttributes.indoor_humidity,
            DeviceAttributes.full_dust,
            DeviceAttributes.total_energy_consumption,
            DeviceAttributes.current_energy_consumption,
            DeviceAttributes.realtime_power,
        ]:
            if attr == DeviceAttributes.prompt_tone:
                self._attributes[DeviceAttributes.prompt_tone] = value
                self.update_all({DeviceAttributes.prompt_tone.value: value})
            elif attr == DeviceAttributes.screen_display:
                message = MessageToggleDisplay(self._protocol_version)
                message.prompt_tone = self._attributes[DeviceAttributes.prompt_tone]
            elif attr in [
                DeviceAttributes.indirect_wind,
                DeviceAttributes.breezeless,
                DeviceAttributes.screen_display_alternate,
            ]:
                message = MessageNewProtocolSet(self._protocol_version)
                setattr(message, str(attr), value)
                message.prompt_tone = self._attributes[DeviceAttributes.prompt_tone]
            elif attr == DeviceAttributes.fresh_air_power:
                if self._fresh_air_version is not None:
                    message = MessageNewProtocolSet(self._protocol_version)
                    setattr(
                        message,
                        str(self._fresh_air_version),
                        [value, self._attributes[DeviceAttributes.fresh_air_fan_speed]],
                    )
            elif attr == DeviceAttributes.fresh_air_mode:
                if value in MideaACDevice._fresh_air_fan_speeds.values():
                    speed = list(MideaACDevice._fresh_air_fan_speeds.keys())[
                        list(MideaACDevice._fresh_air_fan_speeds.values()).index(
                            str(value),
                        )
                    ]
                    fresh_air = (
                        [True, speed]
                        if speed > 0
                        else [
                            False,
                            self._attributes[DeviceAttributes.fresh_air_fan_speed],
                        ]
                    )
                    message = MessageNewProtocolSet(self._protocol_version)
                    setattr(message, str(self._fresh_air_version), fresh_air)
                elif not value:
                    message = MessageNewProtocolSet(self._protocol_version)
                    setattr(
                        message,
                        str(self._fresh_air_version),
                        [False, self._attributes[DeviceAttributes.fresh_air_fan_speed]],
                    )
            elif attr == DeviceAttributes.fresh_air_fan_speed:
                if self._fresh_air_version is not None:
                    message = MessageNewProtocolSet(self._protocol_version)
                    fresh_air = (
                        [True, int(value)]
                        if int(value) > 0
                        else [
                            False,
                            self._attributes[DeviceAttributes.fresh_air_fan_speed],
                        ]
                    )
                    setattr(message, str(self._fresh_air_version), fresh_air)
            elif attr in self._attributes:
                message = self.make_message_set()
                if attr in [
                    DeviceAttributes.boost_mode,
                    DeviceAttributes.sleep_mode,
                    DeviceAttributes.frost_protect,
                    DeviceAttributes.comfort_mode,
                    DeviceAttributes.eco_mode,
                ]:
                    message.boost_mode = False
                    message.sleep_mode = False
                    message.comfort_mode = False
                    message.eco_mode = False
                    message.frost_protect = False
                setattr(message, str(attr), value)
                if attr == DeviceAttributes.mode:
                    setattr(message, str(DeviceAttributes.power.value), True)
        if message is not None:
            self.build_send(message)

    def set_target_temperature(
        self,
        target_temperature: float,
        mode: int | None,
        zone: int | None = None,  # noqa: ARG002
    ) -> None:
        """Midea AC device set target temperature."""
        message: MessageSubProtocolSet | MessageGeneralSet = (
            self.make_message_uniq_set()
        )
        message.target_temperature = target_temperature
        if mode is not None:
            message.power = True
            message.mode = mode
        self.build_send(message)

    def set_swing(self, swing_vertical: bool, swing_horizontal: bool) -> None:
        """Midea AC device set swing."""
        message: MessageGeneralSet = self.make_message_set()
        message.swing_vertical = swing_vertical
        message.swing_horizontal = swing_horizontal
        self.build_send(message)

    def set_customize(self, customize: str) -> None:
        """Midea AC device set custommize."""
        self._temperature_step = self._default_temperature_step
        self._power_analysis_method = self._default_power_analysis_method
        if customize and len(customize) > 0:
            try:
                params = json.loads(customize)
                if params and "temperature_step" in params:
                    self._temperature_step = params.get("temperature_step")
                if params and "power_analysis_method" in params:
                    self._power_analysis_method = params.get("power_analysis_method")
            except Exception:
                _LOGGER.exception("[%s] Set customize error", self.device_id)
            self.update_all({"temperature_step": self._temperature_step})


class MideaAppliance(MideaACDevice):
    """Midea AC appliance."""



"""Midea local x26 message."""

from enum import IntEnum
from typing import Any

from midealocal.const import MAX_BYTE_VALUE
from midealocal.message import MessageBody, MessageRequest, MessageResponse, MessageType

MAX_HEAT_LOW_TEMP = 50


class DeviceMode(IntEnum):
    """Device mode."""

    OFF = 0
    HEAT_HIGH = 1
    HEAT_LOW = 2
    BATH = 3
    BLOW = 4
    VENTILATION = 5
    DRY = 6


class Message26Base(MessageRequest):
    """X26 message base."""

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int,
    ) -> None:
        """Initialize X26 message base."""
        super().__init__(
            device_type=0x26,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError


class MessageQuery(Message26Base):
    """X26 message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize X26 message query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x01,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray([])


class MessageSet(Message26Base):
    """X26 message set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize X26 message set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x01,
        )
        self.fields: dict[str, int] = {}
        self.main_light = False
        self.night_light = False
        self.mode = 0
        self.direction = 0xFD

    def read_field(self, field: str) -> int:
        """X26 message set read field."""
        value = self.fields.get(field, 0)
        return value if value else 0

    @property
    def _body(self) -> bytearray:
        return bytearray(
            [
                1 if self.main_light else 0,
                self.read_field("MAIN_LIGHT_BRIGHTNESS"),
                1 if self.night_light else 0,
                self.read_field("NIGHT_LIGHT_BRIGHTNESS"),
                self.read_field("RADAR_INDUCTION_ENABLE"),
                self.read_field("RADAR_INDUCTION_CLOSING_TIME"),
                self.read_field("LIGHT_INTENSITY_THRESHOLD"),
                self.read_field("RADAR_SENSITIVITY"),
                1 if self.mode in (DeviceMode.HEAT_LOW, DeviceMode.HEAT_HIGH) else 0,
                (
                    0
                    if self.mode not in (DeviceMode.HEAT_LOW, DeviceMode.HEAT_HIGH)
                    else 55
                    if self.mode == DeviceMode.HEAT_HIGH
                    else 30
                ),
                self.read_field("HEATING_SPEED"),
                self.direction,
                1 if self.mode == DeviceMode.BATH else 0,
                self.read_field("BATH_HEATING_TIME"),
                self.read_field("BATH_TEMPERATURE"),
                self.read_field("BATH_SPEED"),
                self.direction,
                1 if self.mode == DeviceMode.VENTILATION else 0,
                self.read_field("VENTILATION_SPEED"),
                self.direction,
                1 if self.mode == DeviceMode.DRY else 0,
                self.read_field("DRYING_TIME"),
                self.read_field("DRYING_TEMPERATURE"),
                self.read_field("DRYING_SPEED"),
                self.direction,
                1 if self.mode == DeviceMode.BLOW else 0,
                self.read_field("BLOWING_SPEED"),
                self.direction,
                self.read_field("DELAY_ENABLE"),
                self.read_field("DELAY_TIME"),
                self.read_field("SOFT_WIND_ENABLE"),
                self.read_field("SOFT_WIND_TIME"),
                self.read_field("SOFT_WIND_TEMPERATURE"),
                self.read_field("SOFT_WIND_SPEED"),
                self.read_field("SOFT_WIND_DIRECTION"),
                self.read_field("WINDLESS_ENABLE"),
                self.read_field("ANION_ENABLE"),
                self.read_field("SMELLY_ENABLE"),
                self.read_field("SMELLY_THRESHOLD"),
            ],
        )


class Message26Body(MessageBody):
    """X26 message body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize X26 message body."""
        super().__init__(body)
        self.fields = self._gen_fields(body)
        self.main_light = self.read_byte(body, 1) > 0
        self.night_light = self.read_byte(body, 3) > 0
        heat_mode = self.read_byte(body, 9) > 0
        heat_temperature = self.read_byte(body, 10)
        heat_direction = self.read_byte(body, 12)
        bath_mode = self.read_byte(body, 13) > 0
        bath_direction = self.read_byte(body, 17)
        ventilation_mode = self.read_byte(body, 18) > 0
        ventilation_direction = self.read_byte(body, 20)
        dry_mode = self.read_byte(body, 21) > 0
        dry_direction = self.read_byte(body, 25)
        blow_mode = self.read_byte(body, 26) > 0
        blow_direction = self.read_byte(body, 28)
        if self.read_byte(body, 31) != MAX_BYTE_VALUE:
            self.current_humidity = self.read_byte(body, 31)
        if self.read_byte(body, 32) != MAX_BYTE_VALUE:
            self.current_radar = self.read_byte(body, 32)
        if self.read_byte(body, 33) != MAX_BYTE_VALUE:
            self.current_temperature = self.read_byte(body, 33)
        self.mode = 0
        self.direction = 0xFD
        if heat_mode:
            if heat_temperature > MAX_HEAT_LOW_TEMP:
                self.mode = 1
            else:
                self.mode = 2
            self.direction = heat_direction
        elif bath_mode:
            self.mode = 3
            self.direction = bath_direction
        elif blow_mode:
            self.mode = 4
            self.direction = blow_direction
        elif ventilation_mode:
            self.mode = 5
            self.direction = ventilation_direction
        elif dry_mode:
            self.mode = 6
            self.direction = dry_direction

    def _gen_fields(self, body: bytearray) -> dict[str, int]:
        fields: dict[str, int] = {}
        fields["MAIN_LIGHT_BRIGHTNESS"] = self.read_byte(body, 2)
        fields["NIGHT_LIGHT_BRIGHTNESS"] = self.read_byte(body, 4)
        fields["RADAR_INDUCTION_ENABLE"] = self.read_byte(body, 5)
        fields["RADAR_INDUCTION_CLOSING_TIME"] = self.read_byte(body, 6)
        fields["LIGHT_INTENSITY_THRESHOLD"] = self.read_byte(body, 7)
        fields["RADAR_SENSITIVITY"] = self.read_byte(body, 8)
        fields["HEATING_SPEED"] = self.read_byte(body, 11)
        fields["BATH_HEATING_TIME"] = self.read_byte(body, 14)
        fields["BATH_TEMPERATURE"] = self.read_byte(body, 15)
        fields["BATH_SPEED"] = self.read_byte(body, 16)
        fields["VENTILATION_SPEED"] = self.read_byte(body, 19)
        fields["DRYING_TIME"] = self.read_byte(body, 22)
        fields["DRYING_TEMPERATURE"] = self.read_byte(body, 23)
        fields["DRYING_SPEED"] = self.read_byte(body, 24)
        fields["BLOWING_SPEED"] = self.read_byte(body, 27)
        fields["DELAY_ENABLE"] = self.read_byte(body, 29)
        fields["DELAY_TIME"] = self.read_byte(body, 30)
        fields["SOFT_WIND_ENABLE"] = self.read_byte(body, 38)
        fields["SOFT_WIND_TIME"] = self.read_byte(body, 39)
        fields["SOFT_WIND_TEMPERATURE"] = self.read_byte(body, 40)
        fields["SOFT_WIND_SPEED"] = self.read_byte(body, 41)
        fields["SOFT_WIND_DIRECTION"] = self.read_byte(body, 42)
        fields["WINDLESS_ENABLE"] = self.read_byte(body, 43)
        fields["ANION_ENABLE"] = self.read_byte(body, 44)
        fields["SMELLY_ENABLE"] = self.read_byte(body, 45)
        fields["SMELLY_THRESHOLD"] = self.read_byte(body, 46)

        return fields


class Message26Response(MessageResponse):
    """X26 message response."""

    def __init__(self, message: bytes) -> None:
        """Initialize X26 message response."""
        super().__init__(bytearray(message))
        if (
            self.message_type
            in [MessageType.set, MessageType.notify1, MessageType.query]
            and self.body_type == 0x01
        ):
            self.set_body(Message26Body(super().body))
        self.fields: dict[str, Any]
        self.set_attr()



"""Midea local A1 device message."""

from enum import IntEnum

from midealocal.crc8 import calculate
from midealocal.message import (
    BodyType,
    MessageBody,
    MessageRequest,
    MessageResponse,
    MessageType,
    NewProtocolMessageBody,
)

MAX_MSG_SERIAL_NUM = 100
MIN_TARGET_HUMIDITY = 35
MIN_FAN_SPEED = 5


class NewProtocolTags(IntEnum):
    """New protocol tags."""

    light = 0x005B


class MessageA1Base(MessageRequest):
    """Message A1 Base."""

    _message_serial = 0

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int,
    ) -> None:
        """Initialize message A1 base."""
        super().__init__(
            device_type=0xA1,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )
        MessageA1Base._message_serial += 1
        if MessageA1Base._message_serial >= MAX_MSG_SERIAL_NUM:
            MessageA1Base._message_serial = 1
        self._message_id = MessageA1Base._message_serial

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError

    @property
    def body(self) -> bytearray:
        """Message A1 base body."""
        body = bytearray([self.body_type]) + self._body + bytearray([self._message_id])
        body.append(calculate(body))
        return body


class MessageQuery(MessageA1Base):
    """Message A1 query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize message A1 query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x41,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray(
            [
                0x81,
                0x00,
                0xFF,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ],
        )


class MessageNewProtocolQuery(MessageA1Base):
    """Message A1 new protocol query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize message A1 new protocol query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0xB1,
        )

    @property
    def _body(self) -> bytearray:
        query_params = [NewProtocolTags.light]
        _body = bytearray([len(query_params)])
        for param in query_params:
            _body.extend([param & 0xFF, param >> 8])
        return _body


class MessageSet(MessageA1Base):
    """Message A1 set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize message A1 set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x48,
        )
        self.power = False
        self.prompt_tone = False
        self.mode = 1
        self.fan_speed = 40
        self.child_lock = False
        self.target_humidity = 40
        self.swing = False
        self.anion = False
        self.water_level_set = 50

    @property
    def _body(self) -> bytearray:
        # byte1, power, prompt_tone
        power = 0x01 if self.power else 0x00
        prompt_tone = 0x40 if self.prompt_tone else 0x00
        # byte2 mode
        mode = self.mode
        # byte3 fan_speed
        fan_speed = self.fan_speed
        # byte7 target_humidity
        target_humidity = self.target_humidity
        # byte8 child_lock
        child_lock = 0x80 if self.child_lock else 0x00
        # byte9 anion
        anion = 0x40 if self.anion else 0x00
        # byte10 swing
        swing = 0x08 if self.swing else 0x00
        # byte 13 water_level_set
        water_level_set = self.water_level_set
        return bytearray(
            [
                power | prompt_tone | 0x02,
                mode,
                fan_speed,
                0x00,
                0x00,
                0x00,
                target_humidity,
                child_lock,
                anion,
                swing,
                0x00,
                0x00,
                water_level_set,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ],
        )


class MessageNewProtocolSet(MessageA1Base):
    """Message A1 new protocol set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize message A1 new protocol set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0xB0,
        )
        self.light: bool | None = None

    @property
    def _body(self) -> bytearray:
        pack_count = 0
        payload = bytearray([0x00])
        if self.light is not None:
            pack_count += 1
            payload.extend(
                NewProtocolMessageBody.pack(
                    param=NewProtocolTags.light,
                    value=bytearray([0x01 if self.light else 0x00]),
                ),
            )
        payload[0] = pack_count
        return payload


class A1GeneralMessageBody(MessageBody):
    """A1 general message body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize A1 general message body."""
        super().__init__(body)
        self.power = (body[1] & 0x01) > 0
        self.mode = body[2] & 0x0F
        self.fan_speed = body[3] & 0x7F
        self.target_humidity = (
            MIN_TARGET_HUMIDITY if (body[7] < MIN_TARGET_HUMIDITY) else body[7]
        )
        self.child_lock = (body[8] & 0x80) > 0
        self.anion = (body[9] & 0x40) > 0
        self.tank = body[10] & 0x7F
        self.water_level_set = body[15]
        self.current_humidity = body[16]
        self.current_temperature = (body[17] - 50) / 2
        self.swing = (body[19] & 0x20) > 0
        if self.fan_speed < MIN_FAN_SPEED:
            self.fan_speed = 1


class A1NewProtocolMessageBody(NewProtocolMessageBody):
    """A1 new protocol message body."""

    def __init__(self, body: bytearray, bt: int) -> None:
        """Initialize A1 new protocol message body."""
        super().__init__(body, bt)
        params = self.parse()
        if NewProtocolTags.light in params:
            self.light = params[NewProtocolTags.light][0] > 0


class MessageA1Response(MessageResponse):
    """A1 message response."""

    def __init__(self, message: bytearray) -> None:
        """Initialize A1 message response."""
        super().__init__(message)
        if self.message_type in [
            MessageType.query,
            MessageType.set,
            MessageType.notify1,
        ]:
            if self.body_type in [BodyType.B0, BodyType.B1, BodyType.B5]:
                self.set_body(A1NewProtocolMessageBody(super().body, self.body_type))
            else:
                self.set_body(A1GeneralMessageBody(super().body))
        elif self.message_type == MessageType.notify2 and self.body_type == BodyType.A0:
            self.set_body(A1GeneralMessageBody(super().body))
        self.set_attr()



"""Midea local FA message."""

from midealocal.message import (
    NONE_VALUE,
    MessageBody,
    MessageRequest,
    MessageResponse,
    MessageType,
    SubBodyType,
)

MAX_FAN_SPEED = 26
TILTING_ANGLE_GET_BYTE = 25
TILTING_ANGLE_SET_BYTE = 24


class MessageFABase(MessageRequest):
    """FA message base."""

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int = NONE_VALUE,
    ) -> None:
        """Initialize the message with protocol version, message type, and body type."""
        super().__init__(
            device_type=0xFA,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError


class MessageQuery(MessageFABase):
    """Message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize the message with protocol version."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
        )

    @property
    def body(self) -> bytearray:
        """Return an empty bytearray."""
        return bytearray([])

    @property
    def _body(self) -> bytearray:
        return bytearray([])


class MessageSet(MessageFABase):
    """Message set."""

    def __init__(self, protocol_version: int, subtype: int) -> None:
        """Initialize the message with protocol version and subtype."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x00,
        )
        self._subtype = subtype
        self.power: bool | None = None
        self.lock: bool | None = None
        self.mode: int | None = None
        self.fan_speed: int | None = None
        self.oscillate: bool | None = None
        self.oscillation_angle: int | None = None
        self.oscillation_mode: int | None = None
        self.tilting_angle: int | None = None

    @property
    def _body(self) -> bytearray:
        if 1 <= self._subtype <= SubBodyType.X0A or self._subtype == SubBodyType.A1:
            _body_return = bytearray(
                [
                    0x00,
                    0x00,
                    0x00,
                    0x80,
                    0x00,
                    0x00,
                    0x00,
                    0x80,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                ],
            )
            if self._subtype != SubBodyType.X0A:
                _body_return[13] = 0xFF
        else:
            _body_return = bytearray(
                [
                    0x00,
                    0x00,
                    0x00,
                    0x80,
                    0x00,
                    0x00,
                    0x00,
                    0x80,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                ],
            )
        if self.power is not None:
            if self.power:
                _body_return[3] = 1
            else:
                _body_return[3] = 0
        if self.lock is not None:
            if self.lock:
                _body_return[2] = 1
            else:
                _body_return[2] = 2
        if self.mode is not None:
            _body_return[3] = 1 | (((self.mode + 1) << 1) & 0x1E)
        if self.fan_speed is not None and 1 <= self.fan_speed <= MAX_FAN_SPEED:
            _body_return[4] = self.fan_speed
        if self.oscillate is not None:
            if self.oscillate:
                _body_return[7] = 1
            else:
                _body_return[7] = 0
        if self.oscillation_angle is not None:
            _body_return[7] = (
                1 | _body_return[7] | ((self.oscillation_angle << 4) & 0x70)
            )
        if self.oscillation_mode is not None:
            _body_return[7] = (
                1 | _body_return[7] | ((self.oscillation_mode << 1) & 0x0E)
            )
        if (
            self.tilting_angle is not None
            and len(_body_return) > TILTING_ANGLE_SET_BYTE
        ):
            _body_return[24] = self.tilting_angle
        return _body_return


class FAGeneralMessageBody(MessageBody):
    """General message body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize the message body."""
        super().__init__(body)
        lock = body[3] & 0x03
        if lock == 1:
            self.child_lock = True
        else:
            self.child_lock = False
        self.power = (body[4] & 0x01) > 0
        mode = (body[4] & 0x1E) >> 1
        if mode > 0:
            self.mode = mode - 1
        fan_speed = body[5]
        if 1 <= fan_speed <= MAX_FAN_SPEED:
            self.fan_speed = fan_speed
        else:
            self.fan_speed = 0
        self.oscillate = (body[8] & 0x01) > 0
        self.oscillation_angle = (body[8] & 0x70) >> 4
        self.oscillation_mode = (body[8] & 0x0E) >> 1
        self.tilting_angle = body[25] if len(body) > TILTING_ANGLE_GET_BYTE else 0


class MessageFAResponse(MessageResponse):
    """FA response message."""

    def __init__(self, message: bytes) -> None:
        """Initialize the message."""
        super().__init__(bytearray(message))
        if self.message_type in [
            MessageType.query,
            MessageType.set,
            MessageType.notify1,
        ]:
            self.set_body(FAGeneralMessageBody(super().body))
        self.set_attr()



"""Midea local BF device."""

import logging
from enum import StrEnum
from typing import Any, ClassVar

from midealocal.device import MideaDevice

from .message import MessageBFResponse, MessageQuery

_LOGGER = logging.getLogger(__name__)


class DeviceAttributes(StrEnum):
    """Midea BF device attributes."""

    door = "door"
    status = "status"
    time_remaining = "time_remaining"
    current_temperature = "current_temperature"
    tank_ejected = "tank_ejected"
    water_change_reminder = "water_change_reminder"
    water_shortage = "water_shortage"


class MideaBFDevice(MideaDevice):
    """Midea BF device."""

    _status: ClassVar[dict[int, str]] = {
        0x01: "PowerSave",
        0x02: "Standby",
        0x03: "Working",
        0x04: "Finished",
        0x05: "Delay",
        0x06: "Paused",
    }

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,  # noqa: ARG002
    ) -> None:
        """Initialize Midea BF device."""
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0xBF,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.door: None,
                DeviceAttributes.status: None,
                DeviceAttributes.time_remaining: None,
                DeviceAttributes.current_temperature: None,
                DeviceAttributes.tank_ejected: None,
                DeviceAttributes.water_change_reminder: None,
                DeviceAttributes.water_shortage: None,
            },
        )

    def build_query(self) -> list[MessageQuery]:
        """Midea BF device build query."""
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg: bytes) -> dict[str, Any]:
        """Midea BF device process message."""
        message = MessageBFResponse(msg)
        _LOGGER.debug("[%s] Received: %s", self.device_id, message)
        new_status = {}
        for status in self._attributes:
            if hasattr(message, str(status)):
                value = getattr(message, str(status))
                if status == DeviceAttributes.status:
                    if value in MideaBFDevice._status:
                        self._attributes[DeviceAttributes.status] = (
                            MideaBFDevice._status.get(value)
                        )
                    else:
                        self._attributes[DeviceAttributes.status] = "Unknown"
                else:
                    self._attributes[status] = value
                new_status[str(status)] = self._attributes[status]
        return new_status

    def set_attribute(self, attr: str, value: bool | int | str) -> None:
        """Midea BF device set attribute."""


class MideaAppliance(MideaBFDevice):
    """Midea BF appliance."""



"""Midea local device."""

import logging
import socket
import threading
import time
from collections.abc import Callable
from enum import IntEnum, StrEnum
from typing import Any

from .exceptions import SocketException
from .message import (
    MessageApplianceResponse,
    MessageQueryAppliance,
    MessageQuestCustom,
    MessageRequest,
    MessageType,
)
from .packet_builder import PacketBuilder
from .security import (
    MSGTYPE_ENCRYPTED_REQUEST,
    MSGTYPE_HANDSHAKE_REQUEST,
    LocalSecurity,
)

MIN_AUTH_RESPONSE = 20
MIN_MSG_LENGTH = 56
MIN_V2_FACTUAL_MSG_LENGTH = 6
RESPONSE_TIMEOUT = 120

_LOGGER = logging.getLogger(__name__)


class DeviceType(IntEnum):
    """Device Type."""

    A0 = 0xA0
    A1 = 0xA1
    AC = 0xAC
    B0 = 0xB0
    B1 = 0xB1
    B3 = 0xB3
    B4 = 0xB4
    B6 = 0xB6
    BF = 0xBF
    C2 = 0xC2
    C3 = 0xC3
    CA = 0xCA
    CC = 0xCC
    CD = 0xCD
    CE = 0xCE
    CF = 0xCF
    DA = 0xDA
    DB = 0xDB
    DC = 0xDC
    E1 = 0xE1
    E2 = 0xE2
    E3 = 0xE3
    E6 = 0xE6
    E8 = 0xE8
    EA = 0xEA
    EC = 0xEC
    ED = 0xED
    FA = 0xFA
    FB = 0xFB
    FC = 0xFC
    FD = 0xFD
    X13 = 0x13
    X26 = 0x26
    X34 = 0x34
    X40 = 0x40


class AuthException(Exception):
    """Authentication exception."""


class ResponseException(Exception):
    """Response exception."""


class RefreshFailed(Exception):
    """Refresh failed exception."""


class DeviceAttributes(StrEnum):
    """Device attributes."""


class ProtocolVersion(IntEnum):
    """Protocol version."""

    V1 = 1
    V2 = 2
    V3 = 3


class ParseMessageResult(IntEnum):
    """Parse message result."""

    SUCCESS = 0
    PADDING = 1
    ERROR = 99


class MideaDevice(threading.Thread):
    """Midea device."""

    def __init__(
        self,
        name: str,
        device_id: int,
        device_type: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        attributes: dict,
    ) -> None:
        """Midea device initialization."""
        threading.Thread.__init__(self)
        self._attributes = attributes or {}
        self._socket: socket.socket | None = None
        self._ip_address = ip_address
        self._port = port
        self._security = LocalSecurity()
        self._token = bytes.fromhex(token)
        self._key = bytes.fromhex(key)
        self._buffer = b""
        self._device_name = name
        self._device_id = device_id
        self._device_type = device_type
        self._protocol = protocol
        self._model = model
        self._subtype = subtype
        self._protocol_version = 0
        self._updates: list[Callable[[dict[str, Any]], None]] = []
        self._unsupported_protocol: list[str] = []
        self._is_run = False
        self._available = True
        self._appliance_query = True
        self._refresh_interval = 30
        self._heartbeat_interval = 10
        self._default_refresh_interval = 30
        self._previous_refresh = 0.0
        self._previous_heartbeat = 0.0
        self.name = self._device_name

    @property
    def available(self) -> bool:
        """Device available."""
        return self._available

    @property
    def device_id(self) -> int:
        """Device ID."""
        return self._device_id

    @property
    def device_type(self) -> int:
        """Device type."""
        return self._device_type

    @property
    def model(self) -> str:
        """Device model."""
        return self._model

    @property
    def subtype(self) -> int:
        """Device subtype."""
        return self._subtype

    @staticmethod
    def fetch_v2_message(msg: bytes) -> tuple[list, bytes]:
        """Fetch V2 message."""
        result = []
        while len(msg) > 0:
            factual_msg_len = len(msg)
            if factual_msg_len < MIN_V2_FACTUAL_MSG_LENGTH:
                break
            alleged_msg_len = msg[4] + (msg[5] << 8)
            if factual_msg_len >= alleged_msg_len:
                result.append(msg[:alleged_msg_len])
                msg = msg[alleged_msg_len:]
            else:
                break
        return result, msg

    def connect(self, refresh_status: bool = True) -> bool:
        """Connect to device."""
        connected = False
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.settimeout(10)
            _LOGGER.debug(
                "[%s] Connecting to %s:%s",
                self._device_id,
                self._ip_address,
                self._port,
            )
            self._socket.connect((self._ip_address, self._port))
            _LOGGER.debug("[%s] Connected", self._device_id)
            if self._protocol == ProtocolVersion.V3:
                self.authenticate()
            _LOGGER.debug("[%s] Authentication success", self._device_id)
            if refresh_status:
                self.refresh_status(wait_response=True)
            connected = True
        except TimeoutError:
            _LOGGER.debug("[%s] Connection timed out", self._device_id)
        except OSError:
            _LOGGER.debug("[%s] Connection error", self._device_id)
        except AuthException:
            _LOGGER.debug("[%s] Authentication failed", self._device_id)
        except ResponseException:
            _LOGGER.debug("[%s] Unexpected response received", self._device_id)
        except RefreshFailed:
            _LOGGER.debug("[%s] Refresh status is timed out", self._device_id)
        except Exception as e:
            file = None
            lineno = None
            if e.__traceback__:
                file = e.__traceback__.tb_frame.f_globals["__file__"]
                lineno = e.__traceback__.tb_lineno
            _LOGGER.exception(
                "[%s] Unknown error : %s, %s",
                self._device_id,
                file,
                lineno,
            )
        self.enable_device(connected)
        return connected

    def authenticate(self) -> None:
        """Authenticate to device. V3 only."""
        request = self._security.encode_8370(self._token, MSGTYPE_HANDSHAKE_REQUEST)
        _LOGGER.debug("[%s] Handshaking", self._device_id)
        if not self._socket:
            raise SocketException
        self._socket.send(request)
        response = self._socket.recv(512)
        if len(response) < MIN_AUTH_RESPONSE:
            raise AuthException
        response = response[8:72]
        self._security.tcp_key(response, self._key)

    def send_message(self, data: bytes) -> None:
        """Send message."""
        if self._protocol == ProtocolVersion.V3:
            self.send_message_v3(data, msg_type=MSGTYPE_ENCRYPTED_REQUEST)
        else:
            self.send_message_v2(data)

    def send_message_v2(self, data: bytes) -> None:
        """Send message V2."""
        if self._socket is not None:
            self._socket.send(data)
        else:
            _LOGGER.debug(
                "[%s] Send failure, device disconnected, data: %s",
                self._device_id,
                data.hex(),
            )

    def send_message_v3(
        self,
        data: bytes,
        msg_type: int = MSGTYPE_ENCRYPTED_REQUEST,
    ) -> None:
        """Send message V3."""
        data = self._security.encode_8370(data, msg_type)
        self.send_message_v2(data)

    def build_send(self, cmd: MessageRequest) -> None:
        """Serialize and send."""
        data = cmd.serialize()
        _LOGGER.debug("[%s] Sending: %s", self._device_id, cmd)
        msg = PacketBuilder(self._device_id, data).finalize()
        self.send_message(msg)

    def refresh_status(self, wait_response: bool = False) -> None:
        """Refresh device status."""
        cmds: list = self.build_query()
        if self._appliance_query:
            cmds = [MessageQueryAppliance(self.device_type), *cmds]
        error_count = 0
        for cmd in cmds:
            if cmd.__class__.__name__ not in self._unsupported_protocol:
                self.build_send(cmd)
                if wait_response:
                    try:
                        while True:
                            if not self._socket:
                                raise SocketException
                            msg = self._socket.recv(512)
                            if len(msg) == 0:
                                raise OSError
                            result = self.parse_message(msg)
                            if result == ParseMessageResult.SUCCESS:
                                break
                            if result == ParseMessageResult.PADDING:
                                continue
                            error_count += 1
                    except TimeoutError:
                        error_count += 1
                        self._unsupported_protocol.append(cmd.__class__.__name__)
                        _LOGGER.debug(
                            "[%s] Does not supports the protocol %s, ignored",
                            self._device_id,
                            cmd.__class__.__name__,
                        )
            else:
                error_count += 1
        if error_count == len(cmds):
            raise RefreshFailed

    def pre_process_message(self, msg: bytearray) -> bool:
        """Pre process message."""
        if msg[9] == MessageType.query_appliance:
            message = MessageApplianceResponse(msg)
            self._appliance_query = False
            _LOGGER.debug("[%s] Received: %s", self._device_id, message)
            self._protocol_version = message.protocol_version
            _LOGGER.debug(
                "[%s] Device protocol version: %s",
                self._device_id,
                self._protocol_version,
            )
            return False
        return True

    def parse_message(self, msg: bytes) -> ParseMessageResult:
        """Parse message."""
        if self._protocol == ProtocolVersion.V3:
            messages, self._buffer = self._security.decode_8370(self._buffer + msg)
        else:
            messages, self._buffer = self.fetch_v2_message(self._buffer + msg)
        if len(messages) == 0:
            return ParseMessageResult.PADDING
        for message in messages:
            if message == b"ERROR":
                return ParseMessageResult.ERROR
            payload_len = message[4] + (message[5] << 8) - 56
            payload_type = message[2] + (message[3] << 8)
            if payload_type in [0x1001, 0x0001]:
                # Heartbeat detected
                pass
            elif len(message) > MIN_MSG_LENGTH:
                cryptographic = bytes(message[40:-16])
                if payload_len % 16 == 0:
                    decrypted: bytearray = self._security.aes_decrypt(cryptographic)
                    try:
                        cont = True
                        if self._appliance_query:
                            cont = self.pre_process_message(decrypted)
                        if cont:
                            status = self.process_message(bytes(decrypted))
                            if len(status) > 0:
                                self.update_all(status)
                            else:
                                _LOGGER.debug(
                                    "[%s] Unidentified protocol",
                                    self._device_id,
                                )

                    except Exception:
                        _LOGGER.exception(
                            "[%s] Error in process message, msg = %s",
                            self._device_id,
                            decrypted.hex(),
                        )
                else:
                    _LOGGER.warning(
                        "[%s] Illegal payload, "
                        "original message = %s, buffer = %s, "
                        "8370 decoded = %s, payload type = %s, "
                        "alleged payload length = %s, factual payload length = %s, ",
                        self._device_id,
                        msg.hex(),
                        self._buffer.hex(),
                        message.hex(),
                        payload_type,
                        payload_len,
                        len(cryptographic),
                    )
            else:
                _LOGGER.warning(
                    "[%s] Illegal message, "
                    "original message = %s, buffer = %s, "
                    "8370 decoded = %s, payload type = %s, "
                    "alleged payload length = %s, message length = %s, ",
                    self._device_id,
                    msg.hex(),
                    self._buffer.hex(),
                    message.hex(),
                    payload_type,
                    payload_len,
                    len(message),
                )
        return ParseMessageResult.SUCCESS

    def build_query(self) -> list:
        """Build query."""
        raise NotImplementedError

    def process_message(self, msg: bytes) -> dict[str, Any]:
        """Process message."""
        raise NotImplementedError

    def send_command(self, cmd_type: int, cmd_body: bytearray) -> None:
        """Send command."""
        cmd = MessageQuestCustom(
            self._device_type,
            self._protocol_version,
            cmd_type,
            cmd_body,
        )
        try:
            self.build_send(cmd)
        except OSError as e:
            _LOGGER.debug(
                "[{%s] Interface send_command failure, %s, "
                "cmd_type: %s, cmd_body: %s",
                self._device_id,
                repr(e),
                cmd_type,
                cmd_body.hex(),
            )

    def send_heartbeat(self) -> None:
        """Send heartbeat."""
        msg = PacketBuilder(self._device_id, bytearray([0x00])).finalize(msg_type=0)
        self.send_message(msg)

    def register_update(self, update: Callable[[dict[str, Any]], None]) -> None:
        """Register update."""
        self._updates.append(update)

    def update_all(self, status: dict[str, Any]) -> None:
        """Update all."""
        _LOGGER.debug("[%s] Status update: %s", self._device_id, status)
        for update in self._updates:
            update(status)

    def enable_device(self, available: bool = True) -> None:
        """Enable device."""
        self._available = available
        status = {"available": available}
        self.update_all(status)

    def open(self) -> None:
        """Open thread."""
        if not self._is_run:
            self._is_run = True
            threading.Thread.start(self)

    def close(self) -> None:
        """Close thread."""
        if self._is_run:
            self._is_run = False
            self.close_socket()

    def close_socket(self) -> None:
        """Close socket."""
        self._unsupported_protocol = []
        self._buffer = b""
        if self._socket:
            self._socket.close()
            self._socket = None

    def set_ip_address(self, ip_address: str) -> None:
        """Set IP address."""
        if self._ip_address != ip_address:
            _LOGGER.debug("[%s] Update IP address to %s", self._device_id, ip_address)
            self._ip_address = ip_address
            self.close_socket()

    def set_refresh_interval(self, refresh_interval: int) -> None:
        """Set refresh interval."""
        self._refresh_interval = refresh_interval

    def _check_refresh(self, now: float) -> None:
        if 0 < self._refresh_interval <= now - self._previous_refresh:
            self.refresh_status()
            self._previous_refresh = now

    def _check_heartbeat(self, now: float) -> None:
        if now - self._previous_heartbeat >= self._heartbeat_interval:
            self.send_heartbeat()
            self._previous_heartbeat = now

    def run(self) -> None:
        """Run loop."""
        while self._is_run:
            while self._socket is None:
                if self.connect(refresh_status=True) is False:
                    self.close_socket()
                    time.sleep(5)
            timeout_counter = 0
            start = time.time()
            self._previous_refresh = start
            self._previous_heartbeat = start
            self._socket.settimeout(1)
            while True:
                try:
                    now = time.time()
                    self._check_refresh(now)
                    self._check_heartbeat(now)
                    msg = self._socket.recv(512)
                    if len(msg) == 0:
                        if self._is_run:
                            _LOGGER.error(
                                "[%s] Socket error - Connection closed by peer",
                                self._device_id,
                            )
                            self.close_socket()
                        break
                    result = self.parse_message(msg)
                    if result == ParseMessageResult.ERROR:
                        _LOGGER.debug("[%s] Message 'ERROR' received", self._device_id)
                        self.close_socket()
                        break
                    if result == ParseMessageResult.SUCCESS:
                        timeout_counter = 0
                except TimeoutError:
                    timeout_counter = timeout_counter + 1
                    if timeout_counter >= RESPONSE_TIMEOUT:
                        _LOGGER.debug("[%s] Heartbeat timed out", self._device_id)
                        self.close_socket()
                        break
                except OSError as e:
                    if self._is_run:
                        _LOGGER.debug("[%s] Socket error %s", self._device_id, repr(e))
                        self.close_socket()
                    break
                except Exception as e:
                    file = None
                    lineno = None
                    if e.__traceback__:
                        file = e.__traceback__.tb_frame.f_globals["__file__"]
                        lineno = e.__traceback__.tb_lineno
                    _LOGGER.exception(
                        "[%s] Unknown error : %s, %s",
                        self._device_id,
                        file,
                        lineno,
                    )

                    self.close_socket()
                    break

    def set_attribute(self, attr: str, value: bool | int | str) -> None:
        """Set attribute."""
        raise NotImplementedError

    def get_attribute(self, attr: str) -> bool | int | str | list[int] | None:
        """Get attribute."""
        return self._attributes.get(attr)

    def set_customize(self, customize: str) -> None:
        """Set customize."""

    @property
    def attributes(self) -> dict[str, Any]:
        """Attributes."""
        ret = {}
        for status in self._attributes:
            ret[str(status)] = self._attributes[status]
        return ret



"""Midea local FC device."""

import json
import logging
from enum import StrEnum
from typing import Any, ClassVar

from midealocal.device import MideaDevice

from .message import MessageFCResponse, MessageQuery, MessageSet

_LOGGER = logging.getLogger(__name__)

STANDBY_DETECT_LENGTH = 2


class DeviceAttributes(StrEnum):
    """Midea FC device attributes."""

    power = "power"
    mode = "mode"
    fan_speed = "fan_speed"
    anion = "anion"
    screen_display = "screen_display"
    detect_mode = "detect_mode"
    pm25 = "pm25"
    tvoc = "tvoc"
    hcho = "hcho"
    child_lock = "child_lock"
    prompt_tone = "prompt_tone"
    filter1_life = "filter1_life"
    filter2_life = "filter2_life"
    standby = "standby"


class MideaFCDevice(MideaDevice):
    """Midea FC device."""

    _modes: ClassVar[dict[int, str]] = {
        0x00: "Standby",
        0x10: "Auto",
        0x20: "Manual",
        0x30: "Sleep",
        0x40: "Fast",
        0x50: "Smoke",
    }
    _speeds: ClassVar[dict[int, str]] = {
        1: "Auto",
        4: "Standby",
        39: "Low",
        59: "Medium",
        80: "High",
    }
    _screen_displays: ClassVar[dict[int, str]] = {0: "Bright", 6: "Dim", 7: "Off"}
    _detect_modes: ClassVar[list[str]] = ["Off", "PM 2.5", "Methanal"]

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,
    ) -> None:
        """Initialize Midea FC device."""
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0xFC,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.power: False,
                DeviceAttributes.mode: None,
                DeviceAttributes.fan_speed: None,
                DeviceAttributes.anion: False,
                DeviceAttributes.standby: False,
                DeviceAttributes.screen_display: None,
                DeviceAttributes.detect_mode: None,
                DeviceAttributes.pm25: None,
                DeviceAttributes.tvoc: None,
                DeviceAttributes.hcho: None,
                DeviceAttributes.child_lock: False,
                DeviceAttributes.prompt_tone: True,
                DeviceAttributes.filter1_life: None,
                DeviceAttributes.filter2_life: None,
            },
        )

        self._standby_detect_default = [40, 20]
        self._standby_detect = self._standby_detect_default
        self.set_customize(customize)

    @property
    def modes(self) -> list[str]:
        """Midea FC device modes."""
        return list(MideaFCDevice._modes.values())

    @property
    def fan_speeds(self) -> list[str]:
        """Midea FC device fan speeds."""
        return list(MideaFCDevice._speeds.values())

    @property
    def screen_displays(self) -> list[str]:
        """Midea FC device screen displays."""
        return list(MideaFCDevice._screen_displays.values())

    @property
    def detect_modes(self) -> list[str]:
        """Midea FC device detect modes."""
        return self._detect_modes

    def build_query(self) -> list[MessageQuery]:
        """Midea FC device build query."""
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg: bytes) -> dict[str, Any]:
        """Midea FC device process message."""
        message = MessageFCResponse(msg)
        _LOGGER.debug("[%s] Received: %s", self.device_id, message)
        new_status = {}
        for status in self._attributes:
            if hasattr(message, str(status)):
                value = getattr(message, str(status))
                if status == DeviceAttributes.mode:
                    if value in MideaFCDevice._modes:
                        self._attributes[status] = MideaFCDevice._modes.get(value)
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.fan_speed:
                    if value in MideaFCDevice._speeds:
                        self._attributes[status] = MideaFCDevice._speeds.get(value)
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.screen_display:
                    if value in MideaFCDevice._screen_displays:
                        self._attributes[status] = MideaFCDevice._screen_displays.get(
                            value,
                        )
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.detect_mode:
                    if value < len(MideaFCDevice._detect_modes):
                        self._attributes[status] = MideaFCDevice._detect_modes[value]
                    else:
                        self._attributes[status] = None
                else:
                    self._attributes[status] = value
                new_status[str(status)] = self._attributes[status]
        return new_status

    def make_message_set(self) -> MessageSet:
        """Midea FC device make message set."""
        message = MessageSet(self._protocol_version)
        message.power = self._attributes[DeviceAttributes.power]
        message.child_lock = self._attributes[DeviceAttributes.child_lock]
        message.prompt_tone = self._attributes[DeviceAttributes.prompt_tone]
        message.anion = self._attributes[DeviceAttributes.anion]
        message.standby = self._attributes[DeviceAttributes.standby]
        message.screen_display = self._attributes[DeviceAttributes.screen_display]
        message.detect_mode = (
            0
            if self._attributes[DeviceAttributes.detect_mode] is None
            else MideaFCDevice._detect_modes.index(
                self._attributes[DeviceAttributes.detect_mode],
            )
        )
        message.mode = (
            0x10
            if self._attributes[DeviceAttributes.mode] is None
            else list(MideaFCDevice._modes.keys())[
                list(MideaFCDevice._modes.values()).index(
                    self._attributes[DeviceAttributes.mode],
                )
            ]
        )
        message.fan_speed = (
            39
            if self._attributes[DeviceAttributes.fan_speed] is None
            else list(MideaFCDevice._speeds.keys())[
                list(MideaFCDevice._speeds.values()).index(
                    self._attributes[DeviceAttributes.fan_speed],
                )
            ]
        )
        message.screen_display = (
            0
            if self._attributes[DeviceAttributes.screen_display] is None
            else list(MideaFCDevice._screen_displays.keys())[
                list(MideaFCDevice._screen_displays.values()).index(
                    self._attributes[DeviceAttributes.screen_display],
                )
            ]
        )
        message.standby_detect = self._standby_detect
        return message

    def set_attribute(self, attr: str, value: bool | int | str) -> None:
        """Midea FC device set attribute."""
        if attr == DeviceAttributes.prompt_tone:
            self._attributes[DeviceAttributes.prompt_tone] = value
            self.update_all({DeviceAttributes.prompt_tone.value: value})
        else:
            message = self.make_message_set()
            if attr == DeviceAttributes.mode:
                if value in MideaFCDevice._modes.values():
                    message.mode = list(MideaFCDevice._modes.keys())[
                        list(MideaFCDevice._modes.values()).index(str(value))
                    ]
            elif attr == DeviceAttributes.fan_speed:
                if value in MideaFCDevice._speeds.values():
                    message.fan_speed = list(MideaFCDevice._speeds.keys())[
                        list(MideaFCDevice._speeds.values()).index(str(value))
                    ]
            elif attr == DeviceAttributes.screen_display:
                if value in MideaFCDevice._screen_displays.values():
                    message.screen_display = list(
                        MideaFCDevice._screen_displays.keys(),
                    )[list(MideaFCDevice._screen_displays.values()).index(str(value))]
                elif not value:
                    message.screen_display = 7
            elif attr == DeviceAttributes.detect_mode:
                if value in MideaFCDevice._detect_modes:
                    message.detect_mode = MideaFCDevice._detect_modes.index(str(value))
                elif not value:
                    message.detect_mode = 0
            else:
                setattr(message, str(attr), value)
            self.build_send(message)

    def set_customize(self, customize: str) -> None:
        """Midea FC device set customize."""
        self._standby_detect = self._standby_detect_default
        if customize and len(customize) > 0:
            try:
                params = json.loads(customize)
                if params and "standby_detect" in params:
                    settings = params.get("standby_detect")
                    if (
                        len(settings) == STANDBY_DETECT_LENGTH
                        and settings[0] > settings[1]
                    ):
                        self._standby_detect = settings
            except Exception:
                _LOGGER.exception("[%s] Set customize error", self.device_id)
            self.update_all({"standby_detect": self._standby_detect})


class MideaAppliance(MideaFCDevice):
    """Midea FC appliance."""



"""Midea local E8 message."""

from enum import IntEnum

from midealocal.message import MessageBody, MessageRequest, MessageResponse, MessageType

MIN_RESPONSE_BODY_LENGTH = 6


class SubCommand(IntEnum):
    """Sub Command."""

    X02 = 0x02
    X04 = 0x04
    X06 = 0x06


class MessageE8Base(MessageRequest):
    """E8 message base."""

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int,
    ) -> None:
        """Initialize E8 message base."""
        super().__init__(
            device_type=0xE8,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError


class MessageQuery(MessageE8Base):
    """E8 message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize E8 message query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0xAA,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray([0x55, 0x00, 0x01, 0x00, 0x00])


class E8MessageBody(MessageBody):
    """E8 message body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize E8 message body."""
        super().__init__(body)
        self.status = body[11]
        self.time_remaining = body[16] * 3600 + body[17] * 60 + body[18]
        self.keep_warm_remaining = body[19] * 3600 + body[20] * 60 + body[21]
        self.working_time = body[28] * 3600 + body[29] * 60 + body[30]
        self.target_temperature = body[39]
        self.current_temperature = body[39]
        self.finished = (body[41] & 0x01) > 0
        self.water_shortage = body[43] > 0


class MessageE8Response(MessageResponse):
    """E8 message response."""

    def __init__(self, message: bytes) -> None:
        """Initialize E8 message response."""
        super().__init__(bytearray(message))
        if len(super().body) > MIN_RESPONSE_BODY_LENGTH:
            sub_cmd = super().body[6]
            if (
                (
                    self.message_type == MessageType.set
                    and sub_cmd in [SubCommand.X02, SubCommand.X04, SubCommand.X06]
                )
                or self.message_type in [MessageType.query, MessageType.notify1]
                and sub_cmd == SubCommand.X02
            ):
                self.set_body(E8MessageBody(super().body))
        self.set_attr()



"""Midea Local E2 device."""

import json
import logging
from enum import IntEnum, StrEnum
from typing import Any

from midealocal.device import MideaDevice

from .message import (
    MessageE2Response,
    MessageNewProtocolSet,
    MessagePower,
    MessageQuery,
    MessageSet,
)


class OldProtocol(StrEnum):
    """Old protocol."""

    auto = "auto"
    true = "true"
    false = "false"


class E2SubType(IntEnum):
    """E2 sub type."""

    T82 = 82
    T85 = 85
    T36353 = 36353


_LOGGER = logging.getLogger(__name__)


class DeviceAttributes(StrEnum):
    """Midea E2 device attributes."""

    power = "power"
    heating = "heating"
    keep_warm = "keep_warm"
    protection = "protection"
    current_temperature = "current_temperature"
    target_temperature = "target_temperature"
    whole_tank_heating = "whole_tank_heating"
    variable_heating = "variable_heating"
    heating_time_remaining = "heating_time_remaining"
    water_consumption = "water_consumption"
    heating_power = "heating_power"


class MideaE2Device(MideaDevice):
    """Midea E2 device."""

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,
    ) -> None:
        """Initialize Midea E2 device."""
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0xE2,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.power: False,
                DeviceAttributes.heating: False,
                DeviceAttributes.keep_warm: False,
                DeviceAttributes.protection: False,
                DeviceAttributes.current_temperature: None,
                DeviceAttributes.target_temperature: 40,
                DeviceAttributes.whole_tank_heating: False,
                DeviceAttributes.variable_heating: False,
                DeviceAttributes.heating_time_remaining: 0,
                DeviceAttributes.water_consumption: None,
                DeviceAttributes.heating_power: None,
            },
        )
        self._default_old_protocol = OldProtocol.auto
        self._old_protocol = self._default_old_protocol
        self.set_customize(customize)

    def _normalize_old_protocol(self, value: str | bool | int) -> OldProtocol:
        if isinstance(value, str):
            return_value = OldProtocol(value)
            if return_value == OldProtocol.auto:
                result = (
                    self.subtype <= E2SubType.T82
                    or self.subtype == E2SubType.T85
                    or self.subtype == E2SubType.T36353
                )
                return_value = OldProtocol.true if result else OldProtocol.false
        if isinstance(value, bool | int):
            return_value = OldProtocol.true if value else OldProtocol.false
        return return_value

    def build_query(self) -> list[MessageQuery]:
        """Midea E2 device build query."""
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg: bytes) -> dict[str, Any]:
        """Midea E2 device process message."""
        message = MessageE2Response(msg)
        _LOGGER.debug("[%s] Received: %s", self.device_id, message)
        new_status = {}
        for status in self._attributes:
            if hasattr(message, str(status)):
                self._attributes[status] = getattr(message, str(status))
                new_status[str(status)] = getattr(message, str(status))
        return new_status

    def make_message_set(self) -> MessageSet:
        """Midea E2 device make message set."""
        message = MessageSet(self._protocol_version)
        message.protection = self._attributes[DeviceAttributes.protection]
        message.whole_tank_heating = self._attributes[
            DeviceAttributes.whole_tank_heating
        ]
        message.target_temperature = self._attributes[
            DeviceAttributes.target_temperature
        ]
        message.variable_heating = self._attributes[DeviceAttributes.variable_heating]
        return message

    def set_attribute(self, attr: str, value: bool | int | str) -> None:
        """Midea E2 device set attribute."""
        message: MessagePower | MessageSet | MessageNewProtocolSet | None = None
        if attr not in [
            DeviceAttributes.heating,
            DeviceAttributes.keep_warm,
            DeviceAttributes.current_temperature,
        ]:
            old_protocol = self._normalize_old_protocol(self._old_protocol)
            if attr == DeviceAttributes.power:
                message = MessagePower(self._protocol_version)
                message.power = bool(value)
            elif old_protocol == OldProtocol.true:
                message = self.make_message_set()
                setattr(message, str(attr), value)
            else:
                message = MessageNewProtocolSet(self._protocol_version)
                setattr(message, str(attr), value)
            self.build_send(message)

    def set_customize(self, customize: str) -> None:
        """Midea E2 device set customize."""
        self._old_protocol = self._default_old_protocol
        if customize and len(customize) > 0:
            try:
                params = json.loads(customize)
                if params and "old_protocol" in params:
                    self._old_protocol = self._normalize_old_protocol(
                        params["old_protocol"],
                    )
            except Exception:
                _LOGGER.exception("[%s] Set customize error", self.device_id)
            self.update_all({"old_protocol": self._old_protocol})


class MideaAppliance(MideaE2Device):
    """Midea E2 appliance."""



"""Midea local FD device."""

import logging
from enum import StrEnum
from typing import Any, ClassVar

from midealocal.device import MideaDevice

from .message import MessageFDResponse, MessageQuery, MessageSet

_LOGGER = logging.getLogger(__name__)

MAX_SUBTYPE_OLD_SPEEDS = 5


class DeviceAttributes(StrEnum):
    """Midea FD device attributes."""

    power = "power"
    fan_speed = "fan_speed"
    prompt_tone = "prompt_tone"
    target_humidity = "target_humidity"
    current_humidity = "current_humidity"
    current_temperature = "current_temperature"
    tank = "tank"
    mode = "mode"
    screen_display = "screen_display"
    disinfect = "disinfect"


class MideaFDDevice(MideaDevice):
    """Midea FD device."""

    _modes: ClassVar[list[str]] = [
        "Manual",
        "Auto",
        "Continuous",
        "Living-Room",
        "Bed-Room",
        "Kitchen",
        "Sleep",
    ]
    _speeds_old: ClassVar[dict[int, str]] = {
        1: "Lowest",
        40: "Low",
        60: "Medium",
        80: "High",
        102: "Auto",
        127: "Off",
    }
    _speeds_new: ClassVar[dict[int, str]] = {
        1: "Lowest",
        39: "Low",
        59: "Medium",
        80: "High",
        101: "Auto",
        127: "Off",
    }
    _screen_displays: ClassVar[dict[int, str]] = {0: "Bright", 6: "Dim", 7: "Off"}
    _detect_modes: ClassVar[list[str]] = ["Off", "PM 2.5", "Methanal"]

    def __init__(
        self,
        name: str,
        device_id: int,
        ip_address: str,
        port: int,
        token: str,
        key: str,
        protocol: int,
        model: str,
        subtype: int,
        customize: str,  # noqa: ARG002
    ) -> None:
        """Initialize Midea FD device."""
        super().__init__(
            name=name,
            device_id=device_id,
            device_type=0xFD,
            ip_address=ip_address,
            port=port,
            token=token,
            key=key,
            protocol=protocol,
            model=model,
            subtype=subtype,
            attributes={
                DeviceAttributes.power: False,
                DeviceAttributes.fan_speed: None,
                DeviceAttributes.prompt_tone: True,
                DeviceAttributes.target_humidity: 60,
                DeviceAttributes.current_humidity: None,
                DeviceAttributes.current_temperature: None,
                DeviceAttributes.tank: 0,
                DeviceAttributes.mode: None,
                DeviceAttributes.screen_display: None,
                DeviceAttributes.disinfect: None,
            },
        )
        if self.subtype > MAX_SUBTYPE_OLD_SPEEDS:
            self._speeds = MideaFDDevice._speeds_new
        else:
            self._speeds = MideaFDDevice._speeds_old

    @property
    def modes(self) -> list[str]:
        """Midea FD device modes."""
        return list(MideaFDDevice._modes)

    @property
    def fan_speeds(self) -> list[str]:
        """Midea FD device fan speeds."""
        return list(self._speeds.values())

    @property
    def screen_displays(self) -> list[str]:
        """Midea FD device screen displays."""
        return list(MideaFDDevice._screen_displays.values())

    @property
    def detect_modes(self) -> list[str]:
        """Midea FD device detect modes."""
        return self._detect_modes

    def build_query(self) -> list[MessageQuery]:
        """Midea FD device build query."""
        return [MessageQuery(self._protocol_version)]

    def process_message(self, msg: bytes) -> dict[str, Any]:
        """Midea FD device process message."""
        message = MessageFDResponse(msg)
        _LOGGER.debug("[%s] Received: %s", self.device_id, message)
        new_status = {}
        for status in self._attributes:
            if hasattr(message, str(status)):
                value = getattr(message, str(status))
                if status == DeviceAttributes.mode:
                    if value <= len(MideaFDDevice._modes):
                        self._attributes[status] = MideaFDDevice._modes[value - 1]
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.fan_speed:
                    if value in self._speeds:
                        self._attributes[status] = self._speeds.get(value)
                    else:
                        self._attributes[status] = None
                elif status == DeviceAttributes.screen_display:
                    if value in MideaFDDevice._screen_displays:
                        self._attributes[status] = MideaFDDevice._screen_displays.get(
                            value,
                        )
                    else:
                        self._attributes[status] = None
                else:
                    self._attributes[status] = value
                new_status[str(status)] = self._attributes[status]
        return new_status

    def make_message_set(self) -> MessageSet:
        """Midea FD device make message set."""
        message = MessageSet(self._protocol_version)
        message.power = self._attributes[DeviceAttributes.power]
        message.prompt_tone = self._attributes[DeviceAttributes.prompt_tone]
        message.screen_display = self._attributes[DeviceAttributes.screen_display]
        message.disinfect = self._attributes[DeviceAttributes.disinfect]
        if self._attributes[DeviceAttributes.mode] in MideaFDDevice._modes:
            message.mode = (
                MideaFDDevice._modes.index(self._attributes[DeviceAttributes.mode]) + 1
            )
        else:
            message.mode = 1
        message.fan_speed = (
            40
            if self._attributes[DeviceAttributes.fan_speed] is None
            else list(self._speeds.keys())[
                list(self._speeds.values()).index(
                    self._attributes[DeviceAttributes.fan_speed],
                )
            ]
        )
        message.screen_display = (
            0
            if self._attributes[DeviceAttributes.screen_display] is None
            else list(MideaFDDevice._screen_displays.keys())[
                list(MideaFDDevice._screen_displays.values()).index(
                    self._attributes[DeviceAttributes.screen_display],
                )
            ]
        )
        return message

    def set_attribute(self, attr: str, value: str | int | bool) -> None:
        """Midea FD device set attribute."""
        if attr == DeviceAttributes.prompt_tone:
            self._attributes[DeviceAttributes.prompt_tone] = value
            self.update_all({DeviceAttributes.prompt_tone.value: value})
        else:
            message = self.make_message_set()
            if attr == DeviceAttributes.mode:
                if value in MideaFDDevice._modes:
                    message.mode = MideaFDDevice._modes.index(str(value)) + 1
            elif attr == DeviceAttributes.fan_speed:
                if value in self._speeds.values():
                    message.fan_speed = list(self._speeds.keys())[
                        list(self._speeds.values()).index(str(value))
                    ]
            elif attr == DeviceAttributes.screen_display:
                if value in MideaFDDevice._screen_displays.values():
                    message.screen_display = list(
                        MideaFDDevice._screen_displays.keys(),
                    )[list(MideaFDDevice._screen_displays.values()).index(str(value))]
                elif not value:
                    message.screen_display = 7
            else:
                setattr(message, str(attr), value)
            self.build_send(message)


class MideaAppliance(MideaFDDevice):
    """Midea FD appliance."""



"""Midea local E2 message."""

from midealocal.message import MessageBody, MessageRequest, MessageResponse, MessageType

HEATING_POWER_BYTE = 34
PROTECTION_BYTE = 22
WATER_CONSUMPTION_BYTE = 25


class MessageE2Base(MessageRequest):
    """E2 message base."""

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int,
    ) -> None:
        """Initialize E2 message base."""
        super().__init__(
            device_type=0xE2,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError


class MessageQuery(MessageE2Base):
    """E2 message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize E2 message query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x01,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray([0x01])


class MessagePower(MessageE2Base):
    """E2 message power."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize E2 message power."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x02,
        )
        self.power = False

    @property
    def _body(self) -> bytearray:
        if self.power:
            self.body_type = 0x01
        else:
            self.body_type = 0x02
        return bytearray([0x01])


class MessageNewProtocolSet(MessageE2Base):
    """E2 message new protocol set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize E2 message new protocol set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x14,
        )
        self.target_temperature: int | None = None
        self.variable_heating: bool | None = None
        self.whole_tank_heating: bool | None = None

    @property
    def _body(self) -> bytearray:
        byte1 = 0x00
        byte2 = 0x00
        if self.target_temperature is not None:
            byte1 = 0x07
            byte2 = int(self.target_temperature) & 0xFF
        elif self.whole_tank_heating is not None:
            byte1 = 0x04
            byte2 = 0x02 if self.whole_tank_heating else 0x01
        elif self.variable_heating is not None:
            byte1 = 0x10
            byte2 = 0x01 if self.variable_heating else 0x00
        return bytearray([byte1, byte2])


class MessageSet(MessageE2Base):
    """E2 message set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize E2 message set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x04,
        )
        self.target_temperature = 0
        self.variable_heating = False
        self.whole_tank_heating = False
        self.protection = False

    @property
    def _body(self) -> bytearray:
        # Byte 4 whole_tank_heating, protection
        protection = 0x04 if self.protection else 0x00
        whole_tank_heating = 0x02 if self.whole_tank_heating else 0x01
        # Byte 5 target_temperature
        target_temperature = self.target_temperature & 0xFF
        # Byte 9 variable_heating
        variable_heating = 0x10 if self.variable_heating else 0x00
        return bytearray(
            [
                0x01,
                0x00,
                0x80,
                whole_tank_heating | protection,
                target_temperature,
                0x00,
                0x00,
                0x00,
                variable_heating,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ],
        )


class E2GeneralMessageBody(MessageBody):
    """E2 message general body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize E2 message general body."""
        super().__init__(body)
        self.power = (body[2] & 0x01) > 0
        self.heating = (body[2] & 0x04) > 0
        self.keep_warm = (body[2] & 0x08) > 0
        self.variable_heating = (body[2] & 0x80) > 0
        self.current_temperature = body[4]
        self.whole_tank_heating = (body[7] & 0x08) > 0
        self.heating_time_remaining = body[9] * 60 + body[10]
        self.target_temperature = body[11]
        self.protection = (
            ((body[22] & 0x02) > 0) if len(body) > PROTECTION_BYTE else False
        )
        if len(body) > WATER_CONSUMPTION_BYTE:
            self.water_consumption = body[24] + (body[25] << 8)
        if len(body) > HEATING_POWER_BYTE:
            self.heating_power = body[34] * 100


class MessageE2Response(MessageResponse):
    """E2 message response."""

    def __init__(self, message: bytes) -> None:
        """Initialize E2 message response."""
        super().__init__(bytearray(message))
        if (
            self.message_type in [MessageType.query, MessageType.notify1]
            and self.body_type == 0x01
        ) or (
            self.message_type == MessageType.set
            and self.body_type in [0x01, 0x02, 0x04, 0x14]
        ):
            self.set_body(E2GeneralMessageBody(super().body))
        self.set_attr()



"""Midea local FC message."""

from midealocal.const import MAX_BYTE_VALUE
from midealocal.crc8 import calculate
from midealocal.message import (
    BodyType,
    MessageBody,
    MessageRequest,
    MessageResponse,
    MessageType,
)

ANION_GET_BYTE = 19
ANION_NOTIFY_BYTE = 10
CHILD_LOCK_GET_BYTE = 8
CHILD_LOCK_NOTIFY_BYTE = 10
DETECT_MODE_GET_BYTE = 29
DETECT_MODE_NOTIFY_BYTE = 22
FILTER1_LIFE_BYTE = 23
FILTER2_LIFE_BYTE = 24
HCHO_GET_BYTE = 38
HCHO_NOTIFY_BYTE = 31
MAX_MSG_SERIAL_NUM = 254
PM25_BYTE = 14
STANDBY_GET_BYTE = 34
STANDBY_NOTIFY_BYTE = 27
STANDBY_VALUE = 0x14
TVOC_BYTE = 15


class MessageFCBase(MessageRequest):
    """FC message base."""

    _message_serial = 0

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int,
    ) -> None:
        """Initialize FC message base."""
        super().__init__(
            device_type=0xFC,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )
        MessageFCBase._message_serial += 1
        if MessageFCBase._message_serial >= MAX_MSG_SERIAL_NUM:
            MessageFCBase._message_serial = 1
        self._message_id = MessageFCBase._message_serial

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError

    @property
    def body(self) -> bytearray:
        """FC message base body."""
        body = bytearray([self.body_type]) + self._body + bytearray([self._message_id])
        body.append(calculate(body))
        return body


class MessageQuery(MessageFCBase):
    """FC message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize FC message query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x41,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray(
            [
                0x00,
                0x00,
                0xFF,
                0x03,
                0x00,
                0x00,
                0x02,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ],
        )


class MessageSet(MessageFCBase):
    """FC message set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize FC message set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x48,
        )
        self.power = False
        self.mode = 0
        self.fan_speed = 0
        self.child_lock = False
        self.prompt_tone = False
        self.anion = False
        self.standby = False
        self.screen_display = 0
        self.detect_mode = 0
        self.standby_detect = [40, 20]

    @property
    def _body(self) -> bytearray:
        # byte1 power
        power = 0x01 if self.power else 0x00
        detect = 0x08 if self.detect_mode > 0 else 0x00
        detect_mode = (self.detect_mode - 1) if self.detect_mode > 0 else 0
        # byte2 mode
        # byte3 fan_speed
        # byte 8 child_lock
        child_lock = 0x80 if self.child_lock else 0x00
        # byte 9 anion
        anion = 0x20 if self.anion else 0x00
        # byte 10 prompt_tone
        prompt_tone = 0x40 if self.prompt_tone else 0x00
        # byte 15/16/17 standby
        if self.standby:
            standby = 0x04
            standby_detect_high = self.standby_detect[0]
            standby_detect_low = self.standby_detect[1]
        else:
            standby = 0x08
            standby_detect_high = 0
            standby_detect_low = 0
        return bytearray(
            [
                power | prompt_tone | detect | 0x02,
                self.mode,
                self.fan_speed,
                0x00,
                0x00,
                0x00,
                0x00,
                child_lock,
                self.screen_display,
                anion,
                0x00,
                0x00,
                0x00,
                detect_mode,
                standby,
                standby_detect_high,
                standby_detect_low,
                0x00,
                0x00,
                0x00,
            ],
        )


class FCGeneralMessageBody(MessageBody):
    """FC message general body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize FC message general body."""
        super().__init__(body)
        self.power = (body[1] & 0x01) > 0
        self.mode = body[2] & 0xF0
        self.fan_speed = body[3] & 0x7F
        self.screen_display = body[9] & 0x07
        self.pm25: int | None = None
        self.tvoc: int | None = None
        self.hcho: int | None = None

        if len(body) > PM25_BYTE and body[14] != MAX_BYTE_VALUE:
            self.pm25 = body[13] + (body[14] << 8)
        if len(body) > TVOC_BYTE and body[15] != MAX_BYTE_VALUE:
            self.tvoc = body[15]
        self.anion = (body[19] & 0x40 > 0) if len(body) > ANION_GET_BYTE else False
        self.standby = (
            ((body[34] & 0xFF) == STANDBY_VALUE)
            if len(body) > STANDBY_GET_BYTE
            else False
        )
        self.child_lock = (
            (body[8] & 0x80 > 0) if len(body) > CHILD_LOCK_GET_BYTE else False
        )
        if len(body) > FILTER1_LIFE_BYTE:
            self.filter1_life = body[23]
        if len(body) > FILTER2_LIFE_BYTE:
            self.filter2_life = body[24]
        if len(body) > DETECT_MODE_GET_BYTE:
            if (body[1] & 0x08) > 0:
                self.detect_mode = body[29] + 1
            else:
                self.detect_mode = 0
        if len(body) > HCHO_GET_BYTE and body[38] != MAX_BYTE_VALUE:
            self.hcho = body[37] + (body[38] << 8)


class FCNotifyMessageBody(MessageBody):
    """FC message notify body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize FC message notify body."""
        super().__init__(body)
        self.power = (body[1] & 0x01) > 0
        self.mode = body[2] & 0xF0
        self.fan_speed = body[3] & 0x7F
        self.screen_display = body[9] & 0x07
        self.pm25: int | None = None
        self.tvoc: int | None = None
        self.hcho: int | None = None

        if len(body) > PM25_BYTE and body[14] != MAX_BYTE_VALUE:
            self.pm25 = body[13] + (body[14] << 8)
        if len(body) > TVOC_BYTE and body[15] != MAX_BYTE_VALUE:
            self.tvoc = body[15]
        self.anion = (body[10] & 0x20 > 0) if len(body) > ANION_NOTIFY_BYTE else False
        self.standby = (
            (body[27] & 0x14 == MAX_BYTE_VALUE)
            if len(body) > STANDBY_NOTIFY_BYTE
            else False
        )
        self.child_lock = (
            (body[10] & 0x10 > 0) if len(body) > CHILD_LOCK_NOTIFY_BYTE else False
        )
        if len(body) > DETECT_MODE_NOTIFY_BYTE:
            if (body[1] & 0x08) > 0:
                self.detect_mode = body[22] + 1
            else:
                self.detect_mode = 0
        if len(body) > HCHO_NOTIFY_BYTE and body[31] != MAX_BYTE_VALUE:
            self.hcho = body[30] + (body[31] << 8)


class MessageFCResponse(MessageResponse):
    """FC message response."""

    def __init__(self, message: bytes) -> None:
        """Initialize FC message response."""
        super().__init__(bytearray(message))
        if self.body_type in [BodyType.B0, BodyType.B1]:
            pass
        elif (
            self.message_type
            in [MessageType.query, MessageType.set, MessageType.notify1]
            and self.body_type == BodyType.C8
        ):
            self.set_body(FCGeneralMessageBody(super().body))
        elif self.message_type == MessageType.notify1 and self.body_type == BodyType.A0:
            self.set_body(FCNotifyMessageBody(super().body))
        self.set_attr()



"""Midea local packet builder."""

from datetime import UTC, datetime
from typing import cast

from .security import LocalSecurity


class PacketBuilder:
    """Packet builder."""

    def __init__(self, device_id: int, command: bytes) -> None:
        """Initialize packet builder."""
        self.command: bytes
        self.security = LocalSecurity()
        # aa20ac00000000000003418100ff03ff000200000000000000000000000006f274
        # Init the packet with the header data.
        self.packet = bytearray(
            [
                # 2 bytes - StaicHeader
                0x5A,
                0x5A,
                # 2 bytes - mMessageType
                0x01,
                0x11,
                # 2 bytes - PacketLenght
                0x00,
                0x00,
                # 2 bytes
                0x20,
                0x00,
                # 4 bytes - MessageId
                0x00,
                0x00,
                0x00,
                0x00,
                # 8 bytes - Date&Time
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                # 6 bytes - mDeviceID
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                # 12 bytes
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
            ],
        )
        self.packet[12:20] = self.packet_time()
        self.packet[20:28] = device_id.to_bytes(8, "little")
        self.command = command

    def finalize(self, msg_type: int = 1) -> bytearray:
        """Finalize packet builder."""
        if msg_type != 1:
            self.packet[3] = 0x10
            self.packet[6] = 0x7B
        else:
            self.packet.extend(self.security.aes_encrypt(self.command))
        # PacketLenght
        self.packet[4:6] = (len(self.packet) + 16).to_bytes(2, "little")
        # Append a basic checksum data(16 bytes) to the packet
        self.packet.extend(self.encode32(self.packet))
        return self.packet

    def encode32(self, data: bytearray) -> bytes:
        """Encode 32."""
        return self.security.encode32_data(data)

    @staticmethod
    def checksum(data: bytes) -> bytes:
        """Packet builder checksum."""
        return cast(bytes, (~sum(data) + 1) & 0xFF)

    @staticmethod
    def packet_time() -> bytearray:
        """Packet builder packet time."""
        t = datetime.now(tz=UTC).strftime("%Y%m%d%H%M%S%f")[:16]
        b = bytearray()
        for i in range(0, len(t), 2):
            d = int(t[i : i + 2])
            b.insert(0, d)
        return b



"""Midea local x13 device."""

from midealocal.message import (
    BodyType,
    MessageBody,
    MessageRequest,
    MessageResponse,
    MessageType,
)

MAX_EFFECT = 5


class Message13Base(MessageRequest):
    """X13 message base."""

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int,
    ) -> None:
        """Initialize X13 message base."""
        super().__init__(
            device_type=0x13,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError


class MessageQuery(Message13Base):
    """X13 message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize X13 message query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x24,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray([0x00, 0x00, 0x00, 0x00])


class MessageSet(Message13Base):
    """X13 message set."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize X13 message set."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.set,
            body_type=0x00,
        )
        self.brightness: int | None = None
        self.color_temperature: int | None = None
        self.effect: int | None = None
        self.power: bool | None = None

    @property
    def _body(self) -> bytearray:
        body_byte = 0x00
        if self.power is not None:
            self.body_type = 0x01
            body_byte = 0x01 if self.power else 0x00
        elif self.effect is not None and self.effect in range(1, 6):
            self.body_type = 0x02
            body_byte = self.effect + 1
        elif self.color_temperature is not None:
            self.body_type = 0x03
            body_byte = self.color_temperature
        elif self.brightness is not None:
            self.body_type = 0x04
            body_byte = self.brightness
        return bytearray([body_byte, 0x00, 0x00, 0x00])


class MessageMainLightBody(MessageBody):
    """X13 message main light body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize X13 message main light body."""
        super().__init__(body)
        self.brightness = self.read_byte(body, 1)
        self.color_temperature = self.read_byte(body, 2)
        self.effect = self.read_byte(body, 3) - 1
        if self.effect > MAX_EFFECT:
            self.effect = 1
        self.power = self.read_byte(body, 8) > 0


class MessageMainLightResponseBody(MessageBody):
    """X13 message main light response body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize X13 message main light response body."""
        super().__init__(body)
        self.control_success: bool = body[1] > 0


class Message13Response(MessageResponse):
    """X13 message response."""

    def __init__(self, message: bytes) -> None:
        """Initialize X13 message response."""
        super().__init__(bytearray(message))
        if self.body_type == BodyType.A4:
            self.set_body(MessageMainLightBody(super().body))
        elif self.message_type == MessageType.set and self.body_type > BodyType.X80:
            self.set_body(MessageMainLightResponseBody(super().body))
        self.control_success: bool
        self.set_attr()



"""Midea local CA message."""

from midealocal.message import (
    BodyType,
    MessageBody,
    MessageRequest,
    MessageResponse,
    MessageType,
)

MIN_CA_GENERAL_BODY_LENGTH = 20
TEMP_POS_LOWER_VALUE = 1
TEMP_POS_UPPER_VALUE = 29
TEMP_NEG_LOWER_VALUE = 49
TEMP_NEG_UPPER_VALUE = 54


class MessageCABase(MessageRequest):
    """CA message base."""

    def __init__(
        self,
        protocol_version: int,
        message_type: int,
        body_type: int,
    ) -> None:
        """Initialize CA message base."""
        super().__init__(
            device_type=0xCA,
            protocol_version=protocol_version,
            message_type=message_type,
            body_type=body_type,
        )

    @property
    def _body(self) -> bytearray:
        raise NotImplementedError


class MessageQuery(MessageCABase):
    """CA message query."""

    def __init__(self, protocol_version: int) -> None:
        """Initialize CA message query."""
        super().__init__(
            protocol_version=protocol_version,
            message_type=MessageType.query,
            body_type=0x00,
        )

    @property
    def _body(self) -> bytearray:
        return bytearray([])


class CAGeneralMessageBody(MessageBody):
    """CA message general body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize CA message general body."""
        super().__init__(body)
        self.refrigerator_setting_temp = body[2] & 0x0F
        self.freezer_setting_temp = -12 - ((body[2] & 0xF0) >> 4)
        flex_zone_setting_temp = body[3]
        right_flex_zone_setting_temp = body[4]

        if TEMP_POS_LOWER_VALUE <= flex_zone_setting_temp <= TEMP_POS_UPPER_VALUE:
            self.flex_zone_setting_temp = flex_zone_setting_temp - 19
        elif TEMP_NEG_LOWER_VALUE <= flex_zone_setting_temp <= TEMP_NEG_UPPER_VALUE:
            self.flex_zone_setting_temp = 30 - flex_zone_setting_temp
        else:
            self.flex_zone_setting_temp = 0
        if TEMP_POS_LOWER_VALUE <= right_flex_zone_setting_temp <= TEMP_POS_UPPER_VALUE:
            self.right_flex_zone_setting_temp = right_flex_zone_setting_temp - 19
        elif (
            TEMP_NEG_LOWER_VALUE <= right_flex_zone_setting_temp <= TEMP_NEG_UPPER_VALUE
        ):
            self.right_flex_zone_setting_temp = 30 - right_flex_zone_setting_temp
        else:
            self.right_flex_zone_setting_temp = 0

        self.energy_consumption = (body[13] << 8) + body[12]
        self.refrigerator_actual_temp = (body[17] - 100) / 2
        self.freezer_actual_temp = (body[18] - 100) / 2
        self.flex_zone_actual_temp = (body[19] - 100) / 2
        self.right_flex_zone_actual_temp = (body[20] - 100) / 2


class CAExceptionMessageBody(MessageBody):
    """CA message exception body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize CA message exception body."""
        super().__init__(body)
        self.refrigerator_door_overtime = (body[1] & 0x01) > 0
        self.freezer_door_overtime = (body[1] & 0x02) > 0
        self.bar_door_overtime = (body[1] & 0x04) > 0
        self.flex_zone_door_overtime = (body[1] & 0x08) > 0


class CANotify00MessageBody(MessageBody):
    """CA message notify00 body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize CA message notify00 body."""
        super().__init__(body)
        self.refrigerator_door = (body[1] & 0x01) > 0
        self.freezer_door = (body[1] & 0x02) > 0
        self.bar_door = (body[1] & 0x04) > 0
        self.flex_zone_door = (body[1] & 0x010) > 0


class CANotify01MessageBody(MessageBody):
    """CA message notify01 body."""

    def __init__(self, body: bytearray) -> None:
        """Initialize CA message notify01 body."""
        super().__init__(body)
        self.refrigerator_setting_temp = body[37]
        self.freezer_setting_temp = -12 - body[38]
        flex_zone_setting_temp = body[39]
        right_flex_zone_setting_temp = body[40]

        if TEMP_POS_LOWER_VALUE <= flex_zone_setting_temp <= TEMP_POS_UPPER_VALUE:
            self.flex_zone_setting_temp = flex_zone_setting_temp - 19
        elif TEMP_NEG_LOWER_VALUE <= flex_zone_setting_temp <= TEMP_NEG_UPPER_VALUE:
            self.flex_zone_setting_temp = 30 - flex_zone_setting_temp
        else:
            self.flex_zone_setting_temp = 0
        if TEMP_POS_LOWER_VALUE <= right_flex_zone_setting_temp <= TEMP_POS_UPPER_VALUE:
            self.right_flex_zone_setting_temp = right_flex_zone_setting_temp - 19
        elif (
            TEMP_NEG_LOWER_VALUE <= right_flex_zone_setting_temp <= TEMP_NEG_UPPER_VALUE
        ):
            self.right_flex_zone_setting_temp = 30 - right_flex_zone_setting_temp
        else:
            self.right_flex_zone_setting_temp = 0


class MessageCAResponse(MessageResponse):
    """CA message response."""

    def __init__(self, message: bytes) -> None:
        """Initialize CA message response."""
        super().__init__(bytearray(message))
        if (
            (
                self.message_type in [MessageType.query, MessageType.set]
                and self.body_type == BodyType.X00
            )
            or (
                self.message_type == MessageType.notify1
                and self.body_type == BodyType.X02
            )
        ) and len(super().body) > MIN_CA_GENERAL_BODY_LENGTH:
            self.set_body(CAGeneralMessageBody(super().body))
        elif (
            self.message_type == MessageType.exception
            and self.body_type == BodyType.X01
        ) or (
            self.message_type == MessageType.query and self.body_type == BodyType.X02
        ):
            self.set_body(CAExceptionMessageBody(super().body))
        elif (
            self.message_type == MessageType.notify1 and self.body_type == BodyType.X00
        ):
            self.set_body(CANotify00MessageBody(super().body))
        elif (
            self.message_type in [MessageType.query, MessageType.notify1]
            and self.body_type == BodyType.X01
        ):
            self.set_body(CANotify01MessageBody(super().body))
        self.set_attr()
